# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict, Iterable, List, Optional, TypeVar, Union

import tensorflow as tf
import torch

from fastestimator.backend.get_gradient import get_gradient
from fastestimator.op.tensorop.tensorop import TensorOp
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import to_list

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)


@traceable()
class GradientOp(TensorOp):
    """Return the gradients of finals w.r.t. inputs.

    Args:
        finals: The tensor(s) to compute gradients from.
        outputs: The key(s) under which to save the gradients.
        inputs: The tensor(s) to compute gradients with respect to, mutually exclusive with `model`.
        model: The model instance to compute gradients with respect to, mutually exclusive with `inputs`.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(self,
                 finals: Union[str, List[str]],
                 outputs: Union[str, List[str]],
                 inputs: Union[None, str, List[str]] = None,
                 model: Union[None, tf.keras.Model, torch.nn.Module] = None,
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None):
        inputs = to_list(inputs)
        finals = to_list(finals)
        outputs = to_list(outputs)
        assert bool(model) != bool(inputs), "Must provide either one of 'inputs' or 'model'"
        if model is None:
            assert len(inputs) == len(finals) == len(outputs), \
                "GradientOp requires the same number of inputs, finals, and outputs"
        else:
            assert isinstance(model, (tf.keras.Model, torch.nn.Module)), "Unrecognized model format"
            assert len(finals) == len(outputs), "GradientOp requires the same number of finals, and outputs"
        inputs.extend(finals)
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        self.model = model
        self.retain_graph = True

    def fe_retain_graph(self, retain: Optional[bool] = None) -> Optional[bool]:
        if retain is not None:
            self.retain_graph = retain
        return self.retain_graph

    def build(self, framework: str, device: Optional[torch.device] = None) -> None:
        self.framework = framework

    def forward(self, data: List[Tensor], state: Dict[str, Any]) -> List[Tensor]:
        results = []
        if self.model is None:
            initials = data[:len(data) // 2]
            finals = data[len(data) // 2:]
            for idx, (initial, final) in enumerate(zip(initials, finals)):
                retain_graph = self.retain_graph or not idx == len(finals) - 1
                results.append(get_gradient(final, initial, tape=state['tape'], retain_graph=retain_graph))
        else:
            finals = data
            if self.framework == "tf":
                trainable_params = self.model.trainable_variables
                for idx, final in enumerate(finals):
                    gradient = get_gradient(final, trainable_params, tape=state['tape'])
                    results.append(gradient)
            elif self.framework == "torch":
                trainable_params = [p for p in self.model.parameters() if p.requires_grad]
                for idx, final in enumerate(finals):
                    # get_gradinet
                    retain_graph = self.retain_graph or not idx == len(finals) - 1
                    gradient = get_gradient(final, trainable_params, retain_graph=retain_graph)
                    results.append(gradient)
            else:
                raise ValueError(f"Unrecognized framework {self.framework}")

        return results



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from collections import ChainMap
from typing import Any, Dict

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
import torch
from PIL import Image

import fastestimator as fe
from fastestimator.dataset.numpy_dataset import NumpyDataset
from fastestimator.op.tensorop.model import ModelOp
from fastestimator.summary import System
from fastestimator.trace import Trace
from fastestimator.util.data import Data


def is_equal(obj1: Any, obj2: Any, assert_type: bool = True, assert_dtype: bool = False) -> bool:
    """Check whether input objects are equal. The object type can be nested iterable (list, tuple, set, dict) and
    with elements such as int, float, np.ndarray, tf.Tensor, tf.Varaible, torch.Tensor

    Args:
        obj1: Input object 1
        obj2: Input object 2
        assert_type: Whether to assert the same data type
        assert_dtype: Whether to assert the same dtype in case of nd.array, tf.Tensor, torch.Tensor

    Returns:
        Boolean of whether those two object are equal
    """
    if assert_type and type(obj1) != type(obj2):
        return False

    if type(obj1) in [list, set, tuple]:
        if len(obj1) != len(obj2):
            return False

        for iter1, iter2 in zip(obj1, obj2):
            if not is_equal(iter1, iter2):
                return False

        return True

    elif type(obj1) == dict:
        if len(obj1) != len(obj2):
            return False

        if obj1.keys() != obj2.keys():
            return False

        for value1, value2 in zip(obj1.values(), obj2.values()):
            if not is_equal(value1, value2):
                return False

        return True

    elif type(obj1) == np.ndarray:
        if assert_dtype and obj1.dtype != obj2.dtype:
            return False
        return np.array_equal(obj1, obj2)

    elif tf.is_tensor(obj1):
        if assert_dtype and obj1.dtype != obj2.dtype:
            return False
        obj1 = obj1.numpy()
        obj2 = obj2.numpy()
        return np.array_equal(obj1, obj2)

    elif isinstance(obj1, torch.Tensor):
        if assert_dtype and obj1.dtype != obj2.dtype:
            return False
        return torch.equal(obj1, obj2)

    else:
        return obj1 == obj2


def one_layer_tf_model() -> tf.keras.Model:
    """Tensorflow Model with one dense layer without activation function.
    * Model input shape: (3,)
    * Model output: (1,)
    * dense layer weight: [1.0, 2.0, 3.0]

    How to feed_forward this model
    ```python
    model = one_layer_tf_model()
    x = tf.constant([[1.0, 1.0, 1.0], [1.0, -1.0, -0.5]])
    b = fe.backend.feed_forward(model, x) # [[6.0], [-2.5]]
    ```

    Returns:
        tf.keras.Model: The model
    """
    inp = tf.keras.layers.Input([3])
    x = tf.keras.layers.Dense(units=1, use_bias=False)(inp)
    model = tf.keras.models.Model(inputs=inp, outputs=x)
    model.layers[1].set_weights([np.array([[1.0], [2.0], [3.0]])])
    return model


def multi_layer_tf_model() -> tf.keras.Model:
    inp = tf.keras.layers.Input([4])
    x = tf.keras.layers.Dense(units=2, use_bias=False, name='fc1')(inp)
    x = tf.keras.layers.Dense(units=1, use_bias=False, name='fc2')(x)
    model = tf.keras.models.Model(inputs=inp, outputs=x)
    model.layers[1].set_weights([np.array([[1.0, 2.0], [2.0, 3.0], [3.0, 4.0], [4.0, 6.0]])])
    model.layers[2].set_weights([np.array([[1.0], [2.0]])])
    return model


class OneLayerTorchModel(torch.nn.Module):
    """Torch Model with one dense layer without activation function.
    * Model input shape: (3,)
    * Model output: (1,)
    * dense layer weight: [1.0, 2.0, 3.0]

    How to feed_forward this model
    ```python
    model = OneLayerTorchModel()
    x = torch.tensor([[1.0, 1.0, 1.0], [1.0, -1.0, -0.5]])
    b = fe.backend.feed_forward(model, x) # [[6.0], [-2.5]]
    ```

    """
    def __init__(self) -> None:
        super().__init__()
        self.fc1 = torch.nn.Linear(3, 1, bias=False)
        self.fc1.weight.data = torch.tensor([[1, 2, 3]], dtype=torch.float32)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        return x


class MultiLayerTorchModel(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.fc1 = torch.nn.Linear(4, 2, bias=False)
        self.fc1.weight.data = torch.tensor([[1, 2, 3, 4], [2, 3, 4, 6]], dtype=torch.float32)
        self.fc2 = torch.nn.Linear(2, 1, bias=False)
        self.fc2.weight.data = torch.tensor([[1, 2]], dtype=torch.float32)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.fc2(x)
        return x


class MockBetaDistribution:
    def __init__(self, framework='tf'):
        self.framework = framework

    def sample(self):
        if self.framework == 'tf':
            return tf.constant(0.5)
        elif self.framework == 'torch':
            return torch.Tensor([0.5])
        else:
            raise ValueError("Unrecognized framework {}".format(self.framework))


class MockUniformDistribution:
    def __init__(self, framework='tf'):
        self.framework = framework

    def sample(self):
        if self.framework == 'tf':
            return tf.constant(0.25)
        elif self.framework == 'torch':
            return torch.Tensor([0.25])
        else:
            raise ValueError("Unrecognized framework {}".format(self.framework))


def sample_system_object():
    x_train = np.random.rand(3, 28, 28, 3)
    y_train = np.random.randint(10, size=(3, ))
    x_eval = np.random.rand(2, 28, 28, 3)
    y_eval = np.random.randint(10, size=(2, ))

    train_data = NumpyDataset({'x': x_train, 'y': y_train})
    eval_data = NumpyDataset({'x': x_eval, 'y': y_eval})
    test_data = eval_data.split(0.5)
    model = fe.build(model_fn=fe.architecture.tensorflow.LeNet, optimizer_fn='adam', model_name='tf')
    pipeline = fe.Pipeline(train_data=train_data, eval_data=eval_data, test_data=test_data, batch_size=1)
    network = fe.Network(ops=[ModelOp(model=model, inputs="x_out", outputs="y_pred")])
    system = System(network=network, pipeline=pipeline, traces=[], total_epochs=10, mode='train')
    return system


def sample_system_object_torch():
    x_train = np.random.rand(3, 28, 28, 3)
    y_train = np.random.randint(10, size=(3, ))
    x_eval = np.random.rand(2, 28, 28, 3)
    y_eval = np.random.randint(10, size=(2, ))

    train_data = NumpyDataset({'x': x_train, 'y': y_train})
    eval_data = NumpyDataset({'x': x_eval, 'y': y_eval})
    test_data = eval_data.split(0.5)
    model = fe.build(model_fn=fe.architecture.pytorch.LeNet, optimizer_fn='adam', model_name='torch')
    pipeline = fe.Pipeline(train_data=train_data, eval_data=eval_data, test_data=test_data, batch_size=1)
    network = fe.Network(ops=[ModelOp(model=model, inputs="x_out", outputs="y_pred")])
    system = System(network=network, pipeline=pipeline, traces=[], total_epochs=10, mode='train')
    return system


def check_img_similar(img1: np.ndarray, img2: np.ndarray, ptol: int = 3, ntol: float = 0.01) -> bool:
    """Check whether img1 and img2 array are similar based on pixel to pixel comparision
    Args:
        img1: Image 1
        img2: Image 2
        ptol: Pixel value tolerance
        ntol: Number of pixel difference tolerace rate

    Returns:
        Boolean of whether the images are similar
    """
    if img1.shape == img2.shape:
        diff = np.abs(img1.astype(np.float32) - img2.astype(np.float32))
        n_pixel_diff = diff[diff > ptol].size
        if n_pixel_diff < img1.size * ntol:
            return True
        else:
            return False
    return False


def img_to_rgb_array(path: str) -> np.ndarray:
    """Read png file to numpy array (RGB)

    Args:
        path: Image path

    Returns:
        Image numpy array
    """
    return np.asarray(Image.open(path).convert('RGB'))


def fig_to_rgb_array(fig: plt.Figure) -> np.ndarray:
    """Convert image in plt.Figure to numpy array

    Args:
        fig: Input figure object

    Returns:
        Image array
    """
    fig.canvas.draw()
    buf = fig.canvas.tostring_rgb()
    ncols, nrows = fig.canvas.get_width_height()
    return np.frombuffer(buf, dtype=np.uint8).reshape(nrows, ncols, 3)


class TraceRun:
    """Class to simulate the trace calling protocol.

    This serve for testing purpose without using estimator class.

    Args:
        trace: Target trace to run.
        batch: Batch data from pipepline.
        prediction: Batch data from network.
    """
    def __init__(self, trace: Trace, batch: Dict[str, Any], prediction: Dict[str, Any]):
        self.trace = trace
        self.batch = batch
        self.prediction = prediction
        self.data_on_begin = None
        self.data_on_end = None
        self.data_on_epoch_begin = None
        self.data_on_epoch_end = None
        self.data_on_batch_begin = None
        self.data_on_batch_end = None

    def run_trace(self) -> None:
        system = sample_system_object()
        self.trace.system = system

        self.data_on_begin = Data()
        self.trace.on_begin(self.data_on_begin)

        self.data_on_epoch_begin = Data()
        self.trace.on_epoch_begin(self.data_on_epoch_begin)

        self.data_on_batch_begin = Data(self.batch)
        self.trace.on_batch_begin(self.data_on_batch_begin)

        self.data_on_batch_end = Data(ChainMap(self.prediction, self.batch))
        self.trace.on_batch_end(self.data_on_batch_end)

        self.data_on_epoch_end = Data()
        self.trace.on_epoch_end(self.data_on_epoch_end)

        self.data_on_end = Data()
        self.trace.on_end(self.data_on_end)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np

from fastestimator.op.numpyop.univariate.hadamard import Hadamard
from fastestimator.test.unittest_util import is_equal


class TestHadamard(unittest.TestCase):
    def test_4class(self):
        tohadamard = Hadamard(inputs='y', outputs='y', n_classes=4)
        output = tohadamard.forward(data=[np.array([3.0, 2.0, 2.0, 2.0, 0.0])], state={})[0]
        self.assertTrue(
            is_equal(
                output,
                np.array([[1., -1., -1., 1.], [-1., 1., -1., -1.], [-1., 1., -1., -1.], [-1., 1., -1., -1.],
                          [-1., 1., 1., 1.]])))

    def test_4class_single_input(self):
        tohadamard = Hadamard(inputs='y', outputs='y', n_classes=4)
        output = tohadamard.forward(data=[0], state={})[0]
        self.assertTrue(is_equal(output, np.array([-1., 1., 1., 1.])))

    def test_10class_16code(self):
        tohadamard = Hadamard(inputs='y', outputs='y', n_classes=10, code_length=16)
        output = tohadamard.forward(data=[np.array([[0], [1], [9], [0], [4]])], state={})[0]
        self.assertTrue(
            is_equal(
                output,
                np.array([[-1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.,
                           1.], [1., -1., 1., -1., 1., -1., 1., -1., 1., -1., 1., -1., 1., -1., 1.,
                                 -1.], [1., -1., 1., -1., 1., -1., 1., -1., -1., 1., -1., 1., -1., 1., -1.,
                                        1.], [-1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.],
                          [-1., 1., 1., 1., -1., -1., -1., -1., 1., 1., 1., 1., -1., -1., -1., -1.]])))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe


class TestTensorPow(unittest.TestCase):
    def test_np_input_pow_gt_1(self):
        n = np.array([[1, 4, 6], [2.3, 0.5, 0]])
        target = np.array([[1, 8.44485063e+01, 3.09089322e+02], [1.43723927e+01, 1.08818820e-01, 0]])
        b = fe.backend.tensor_pow(n, 3.2)
        self.assertTrue(np.allclose(b, target))

    def test_np_input_pow_lt_1(self):
        n = np.array([[1, 4, 6], [2.3, 0.5, 0]])
        target = np.array([[1, 1.33792755, 1.45683968], [1.1911401, 0.86453723, 0]])
        b = fe.backend.tensor_pow(n, 0.21)
        self.assertTrue(np.allclose(b, target))

    def test_tf_input_pow_gt_1(self):
        n = tf.convert_to_tensor([[1, 4, 6], [2.3, 0.5, 0]])
        target = tf.convert_to_tensor([[1, 8.44485063e+01, 3.09089322e+02], [1.43723927e+01, 1.08818820e-01, 0]])
        b = fe.backend.tensor_pow(n, 3.2)
        self.assertTrue(np.allclose(b, target))

    def test_tf_input_pow_lt_1(self):
        n = tf.convert_to_tensor([[1, 4, 6], [2.3, 0.5, 0]])
        target = tf.convert_to_tensor([[1, 1.33792755, 1.45683968], [1.1911401, 0.86453723, 0]])
        b = fe.backend.tensor_pow(n, 0.21)
        self.assertTrue(np.allclose(b, target))

    def test_torch_input_pow_gt_1(self):
        n = torch.tensor([[1, 4, 6], [2.3, 0.5, 0]])
        target = torch.tensor([[1, 8.44485063e+01, 3.09089322e+02], [1.43723927e+01, 1.08818820e-01, 0]])
        b = fe.backend.tensor_pow(n, 3.2)
        self.assertTrue(np.allclose(b, target))

    def test_torch_input_pow_lt_1(self):
        n = torch.tensor([[1, 4, 6], [2.3, 0.5, 0]])
        target = torch.tensor([[1, 1.33792755, 1.45683968], [1.1911401, 0.86453723, 0]])
        b = fe.backend.tensor_pow(n, 0.21)
        self.assertTrue(np.allclose(b, target))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.test.unittest_util import is_equal


class TestConcat(unittest.TestCase):
    def test_concat_np_input_axis_0(self):
        t = [np.array([[0, 1]]), np.array([[2, 3]]), np.array([[4, 5]])]
        obj1 = fe.backend.concat(t, axis=0)
        obj2 = np.array([[0, 1], [2, 3], [4, 5]])
        self.assertTrue(is_equal(obj1, obj2))

    def test_concat_np_input_axis_1(self):
        t = [np.array([[0, 1]]), np.array([[2, 3]]), np.array([[4, 5]])]
        obj1 = fe.backend.concat(t, axis=1)
        obj2 = np.array([[0, 1, 2, 3, 4, 5]])
        self.assertTrue(is_equal(obj1, obj2))

    def test_concat_tf_input_axis_0(self):
        t = [tf.constant([[0, 1]]), tf.constant([[2, 3]]), tf.constant([[4, 5]])]
        obj1 = fe.backend.concat(t, axis=0)
        obj2 = tf.constant([[0, 1], [2, 3], [4, 5]])
        self.assertTrue(is_equal(obj1, obj2))

    def test_concat_tf_input_axis_1(self):
        t = [tf.constant([[0, 1]]), tf.constant([[2, 3]]), tf.constant([[4, 5]])]
        obj1 = fe.backend.concat(t, axis=1)
        obj2 = tf.constant([[0, 1, 2, 3, 4, 5]])
        self.assertTrue(is_equal(obj1, obj2))

    def test_concat_torch_input_axis_0(self):
        t = [torch.tensor([[0, 1]]), torch.tensor([[2, 3]]), torch.tensor([[4, 5]])]
        obj1 = fe.backend.concat(t, axis=0)
        obj2 = torch.tensor([[0, 1], [2, 3], [4, 5]])
        self.assertTrue(is_equal(obj1, obj2))

    def test_concat_torch_input_axis_1(self):
        t = [torch.tensor([[0, 1]]), torch.tensor([[2, 3]]), torch.tensor([[4, 5]])]
        obj1 = fe.backend.concat(t, axis=1)
        obj2 = torch.tensor([[0, 1, 2, 3, 4, 5]])
        self.assertTrue(is_equal(obj1, obj2))



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import tempfile
import unittest

from fastestimator.search.search import Search


class DummySearch(Search):
    def _fit(self):
        pass


class TestSearch(unittest.TestCase):
    def test_scoring_wo_index(self):
        with self.assertRaises(AssertionError):
            DummySearch(score_fn=lambda x: x)

    def test_index_increase(self):
        search = DummySearch(score_fn=lambda search_idx, x: x)
        search.evaluate(x=1)
        search.evaluate(x=2)
        self.assertEqual(search.search_idx, 2)

    def test_caching(self):
        search = DummySearch(score_fn=lambda search_idx, x: x)
        search.evaluate(x=1)
        search.evaluate(x=1)  # second time should not propagate the index
        self.assertEqual(search.search_idx, 1)

    def test_best_param_max(self):
        search = DummySearch(score_fn=lambda search_idx, x: x)
        search.evaluate(x=1)
        search.evaluate(x=2)
        self.assertEqual(search.get_best_results(), ({"search_idx": 2, "x": 2}, 2))

    def test_best_param_min(self):
        search = DummySearch(score_fn=lambda search_idx, x: x, best_mode="min")
        search.evaluate(x=1)
        search.evaluate(x=2)
        self.assertEqual(search.get_best_results(), ({"search_idx": 1, "x": 1}, 1))

    def test_get_state(self):
        search = DummySearch(score_fn=lambda search_idx, x: x)
        search.evaluate(x=1)
        search.evaluate(x=2)
        search.evaluate(x=1)
        search.evaluate(x=5)
        state = search._get_state()
        self.assertEqual(len(state["search_results"]), 3)

    def test_save(self):
        search = DummySearch(score_fn=lambda search_idx, x: x, name="unicorn")
        save_dir = tempfile.mkdtemp()
        search.save(save_dir=save_dir)
        self.assertTrue(os.path.exists(os.path.join(save_dir, "unicorn.json")))

    def test_load(self):
        search = DummySearch(score_fn=lambda search_idx, x: x, name="ultraman")
        search.evaluate(x=1)
        search.evaluate(x=2)
        save_dir = tempfile.mkdtemp()
        search.save(save_dir=save_dir)
        # now to load
        search2 = DummySearch(score_fn=lambda search_idx, x: x, name="ultraman")
        search2.load(load_dir=save_dir)
        search2.evaluate(x=1)
        self.assertEqual(search2.search_idx, 2)

    def test_load_empty_dir(self):
        search = DummySearch(score_fn=lambda search_idx, x: x, name="pikachu")
        with self.assertRaises(ValueError):
            search.load(tempfile.mkdtemp())

    def test_load_empty_dir_not_exists_ok(self):
        search = DummySearch(score_fn=lambda search_idx, x: x, name="pikachu")
        search.load(tempfile.mkdtemp(), not_exist_ok=True)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import tensorflow as tf
import torch

from fastestimator.op.tensorop import LambdaOp
from fastestimator.op.tensorop.meta import OneOf


class TestOneOf(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.output_shape = (28, 28, 3)
        cls.single_input_tf = [tf.random.uniform(shape=(28, 28, 3))]
        cls.multi_input_tf = [tf.random.uniform(shape=(28, 28, 3)), tf.random.uniform(shape=(28, 28, 3))]
        cls.single_input_torch = [torch.ones(size=(28, 28, 3))]
        cls.multi_input_torch = [torch.ones(size=(28, 28, 3)), torch.ones(size=(28, 28, 3))]

    def test_single_input_tf(self):
        a = LambdaOp(inputs='x', outputs='x', fn=lambda x: x + 1)
        b = LambdaOp(inputs='x', outputs='x', fn=lambda x: x + 5)
        oneof = OneOf(a, b)
        oneof.build('tf')
        output = oneof.forward(data=self.single_input_tf, state={})
        with self.subTest('Check output type'):
            self.assertTrue(tf.is_tensor(output))
        with self.subTest('Check output image shape'):
            self.assertEqual(output.shape, self.output_shape)

    def test_multi_input_tf(self):
        a = LambdaOp(inputs=['x', 'y'], outputs=['y', 'z'], fn=lambda x, y: [x + y, x - y])
        b = LambdaOp(inputs=['x', 'y'], outputs=['y', 'z'], fn=lambda x, y: [x * y, x + y])
        c = LambdaOp(inputs=['x', 'y'], outputs=['y', 'z'], fn=lambda x, y: [y, x])
        oneof = OneOf(a, b, c)
        oneof.build('tf')
        output = oneof.forward(data=self.multi_input_tf, state={})
        with self.subTest('Check output type'):
            self.assertEqual(type(output), list)
        with self.subTest('Check output list length'):
            self.assertEqual(len(output), 2)
        for img_output in output:
            with self.subTest('Check output image shape'):
                self.assertEqual(img_output.shape, self.output_shape)

    @tf.function
    def test_single_input_tf_static(self):
        a = LambdaOp(inputs='x', outputs='x', fn=lambda x: x + 1)
        b = LambdaOp(inputs='x', outputs='x', fn=lambda x: x + 5)
        oneof = OneOf(a, b)
        oneof.build('tf')
        output = oneof.forward(data=self.single_input_tf, state={})
        with self.subTest('Check output type'):
            self.assertTrue(tf.is_tensor(output))
        with self.subTest('Check output image shape'):
            self.assertEqual(output.shape, self.output_shape)

    @tf.function
    def test_multi_input_tf_static(self):
        a = LambdaOp(inputs=['x', 'y'], outputs=['y', 'z'], fn=lambda x, y: [x + y, x - y])
        b = LambdaOp(inputs=['x', 'y'], outputs=['y', 'z'], fn=lambda x, y: [x * y, x + y])
        c = LambdaOp(inputs=['x', 'y'], outputs=['y', 'z'], fn=lambda x, y: [y, x])
        oneof = OneOf(a, b, c)
        oneof.build('tf')
        output = oneof.forward(data=self.multi_input_tf, state={})
        with self.subTest('Check output type'):
            self.assertEqual(type(output), list)
        with self.subTest('Check output list length'):
            self.assertEqual(len(output), 2)
        for img_output in output:
            with self.subTest('Check output image shape'):
                self.assertEqual(img_output.shape, self.output_shape)

    def test_single_input_torch(self):
        a = LambdaOp(inputs='x', outputs='x', fn=lambda x: x + 1)
        b = LambdaOp(inputs='x', outputs='x', fn=lambda x: x + 5)
        oneof = OneOf(a, b)
        oneof.build('torch')
        output = oneof.forward(data=self.single_input_torch, state={})
        with self.subTest('Check output type'):
            self.assertEqual(type(output), torch.Tensor)
        with self.subTest('Check output image shape'):
            self.assertEqual(output.shape, self.output_shape)

    def test_multi_input_torch(self):
        a = LambdaOp(inputs=['x', 'y'], outputs=['y', 'z'], fn=lambda x, y: [x + y, x - y])
        b = LambdaOp(inputs=['x', 'y'], outputs=['y', 'z'], fn=lambda x, y: [x * y, x + y])
        c = LambdaOp(inputs=['x', 'y'], outputs=['y', 'z'], fn=lambda x, y: [y, x])
        oneof = OneOf(a, b, c)
        oneof.build('torch')
        output = oneof.forward(data=self.multi_input_torch, state={})
        with self.subTest('Check output type'):
            self.assertEqual(type(output), list)
        with self.subTest('Check output list length'):
            self.assertEqual(len(output), 2)
        for img_output in output:
            with self.subTest('Check output image shape'):
                self.assertEqual(img_output.shape, self.output_shape)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

from fastestimator.op.tensorop.loss import CrossEntropy


class TestCrossEntropy(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # binary ce
        cls.tf_true_binary = tf.constant([[1.0], [2.0], [4.0]])
        cls.tf_pred_binary = tf.constant([[1.0], [3.0], [4.5]])
        cls.tf_binary_weights = {2: 2.0, 4: 3.0}
        # torch binary ce
        cls.torch_true_binary = torch.tensor([[1], [0], [1], [0]])
        cls.torch_pred_binary = torch.tensor([[0.9], [0.3], [0.8], [0.1]])
        cls.torch_binary_weights = {1: 2.0}
        # categorical ce
        cls.tf_true_cat = tf.constant([[0, 1, 0], [1, 0, 0], [0, 0, 1]])
        cls.tf_pred_cat = tf.constant([[0.1, 0.8, 0.1], [0.9, 0.05, 0.05], [0.1, 0.2, 0.7]])
        cls.tf_cat_weights = {1: 2.0, 2: 3.0}
        # sparse categorical ce
        cls.tf_true_sparse = tf.constant([[0], [1], [0]])
        cls.tf_pred_sparse = tf.constant([[0.1, 0.8, 0.1], [0.9, 0.05, 0.05], [0.1, 0.2, 0.7]])
        cls.tf_sparse_weights = {1: 2.0, 2: 3.0}

    def test_binary_crossentropy(self):
        ce = CrossEntropy(inputs='x', outputs='x')
        output = ce.forward(data=[self.tf_pred_binary, self.tf_true_binary], state={})
        self.assertTrue(np.allclose(output.numpy(), -20.444319))

    def test_binary_crossentropy_weights(self):
        ce = CrossEntropy(inputs='x', outputs='x', class_weights=self.tf_binary_weights)
        ce.build('tf')
        output = ce.forward(data=[self.tf_pred_binary, self.tf_true_binary], state={})
        self.assertTrue(np.allclose(output.numpy(), -56.221874))

    def test_categorical_crossentropy(self):
        ce = CrossEntropy(inputs='x', outputs='x')
        output = ce.forward(data=[self.tf_pred_cat, self.tf_true_cat], state={})
        self.assertTrue(np.allclose(output.numpy(), 0.22839302))

    def test_categorical_crossentropy_weights(self):
        ce = CrossEntropy(inputs='x', outputs='x', class_weights=self.tf_cat_weights)
        ce.build('tf')
        output = ce.forward(data=[self.tf_pred_cat, self.tf_true_cat], state={})
        self.assertTrue(np.allclose(output.numpy(), 0.54055756))

    def test_sparse_categorical_crossentropy(self):
        ce = CrossEntropy(inputs='x', outputs='x')
        output = ce.forward(data=[self.tf_pred_sparse, self.tf_true_sparse], state={})
        self.assertTrue(np.allclose(output.numpy(), 2.5336342))

    def test_sparse_categorical_crossentropy_weights(self):
        ce = CrossEntropy(inputs='x', outputs='x', class_weights=self.tf_sparse_weights)
        ce.build('tf')
        output = ce.forward(data=[self.tf_pred_sparse, self.tf_true_sparse], state={})
        self.assertTrue(np.allclose(output.numpy(), 3.532212))

    def test_torch_input(self):
        ce = CrossEntropy(inputs='x', outputs='x')
        output = ce.forward(data=[self.torch_pred_binary, self.torch_true_binary], state={})
        self.assertTrue(np.allclose(output.detach().numpy(), 0.1976349))

    def test_torch_input_weights(self):
        ce = CrossEntropy(inputs='x', outputs='x', class_weights=self.torch_binary_weights)
        ce.build('torch')
        output = ce.forward(data=[self.torch_pred_binary, self.torch_true_binary], state={})
        self.assertTrue(np.allclose(output.detach().numpy(), 0.2797609))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from fastestimator.op.tensorop import augmentation, gradient, loss, meta, model
from fastestimator.op.tensorop.argmax import Argmax
from fastestimator.op.tensorop.average import Average
from fastestimator.op.tensorop.gather import Gather
from fastestimator.op.tensorop.reshape import Reshape
from fastestimator.op.tensorop.tensorop import LambdaOp, TensorOp
from fastestimator.op.tensorop.un_hadamard import UnHadamard



#  Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ==============================================================================
import argparse
import os
from typing import Any, Dict, List, Optional, Sequence, Union

from fastestimator.util.util import parse_string_to_python


class SaveAction(argparse.Action):
    """A customized save action for use with argparse.

    A custom save action which is used to populate a secondary variable inside of an exclusive group. Used if this file
    is invoked directly during argument parsing.

    This class is intentionally not @traceable.

    Args:
        option_strings: A list of command-line option strings which should be associated with this action.
        dest: The name of the attribute to hold the created object(s).
        nargs: The number of command line arguments to be consumed.
        **kwargs: Pass-through keyword arguments.
    """
    def __init__(self,
                 option_strings: Sequence[str],
                 dest: str,
                 nargs: Union[int, str, None] = '?',
                 **kwargs: Dict[str, Any]) -> None:
        if '?' != nargs:
            raise ValueError("nargs must be \'?\'")
        super().__init__(option_strings, dest, nargs, **kwargs)

    def __call__(self,
                 parser: argparse.ArgumentParser,
                 namespace: argparse.Namespace,
                 values: Optional[str],
                 option_string: Optional[str] = None) -> None:
        """Invokes the save action, writing two values into the namespace.

        Args:
            parser: The active argument parser (ignored by this implementation).
            namespace: The current namespace to be written to.
            values: The value to write into the namespace.
            option_string: An option_string (ignored by this implementation).
        """
        setattr(namespace, self.dest, True)
        setattr(namespace, self.dest + '_dir', values if values is None else os.path.join(values, ''))


def parse_cli_to_dictionary(input_list: List[str]) -> Dict[str, Any]:
    """Convert a list of strings into a dictionary with python objects as values.

    ```python
    a = parse_cli_to_dictionary(["--epochs", "5", "--test", "this", "--lr", "0.74"])
    # {'epochs': 5, 'test': 'this', 'lr': 0.74}
    ```

    Args:
        input_list: A list of input strings from the cli.

    Returns:
        A dictionary constructed from the `input_list`, with values converted to python objects where applicable.
    """
    result = {}
    if input_list is None:
        return result
    key = ""
    val = ""
    idx = 0
    while idx < len(input_list):
        if input_list[idx].startswith("--"):
            if len(key) > 0:
                result[key] = parse_string_to_python(val)
            val = ""
            key = input_list[idx].strip('--')
        else:
            val += input_list[idx]
        idx += 1
    if len(key) > 0:
        result[key] = parse_string_to_python(val)
    return result



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import re
import statistics
from collections import defaultdict
from typing import Any, Dict, List, NamedTuple, Optional, Union

from fastestimator.util.traceability_util import FeSummaryTable


class ValWithError(NamedTuple):
    """A class to record values with error bars (for special visualization in the logger).
    """
    y_min: float
    y: float
    y_max: float

    def __str__(self):
        return f"({self.y_min}, {self.y}, {self.y_max})"


class Summary:
    """A summary object that records training history.

    This class is intentionally not @traceable.

    Args:
        name: Name of the experiment. If None then experiment results will be ignored.
        system_config: A description of the initialization parameters defining the estimator associated with this
            experiment.
    """
    def __init__(self, name: Optional[str], system_config: Optional[List[FeSummaryTable]] = None) -> None:
        self.name = name
        self.system_config = system_config
        self.history = defaultdict(lambda: defaultdict(dict))  # {mode: {key: {step: value}}}

    def merge(self, other: 'Summary'):
        """Merge another `Summary` into this one.

        Args:
            other: Other `summary` object to be merged.
        """
        for mode, sub in other.history.items():
            for key, val in sub.items():
                self.history[mode][key].update(val)

    def __bool__(self) -> bool:
        """Whether training history should be recorded.

        Returns:
            True iff this `Summary` has a non-None name.
        """
        return bool(self.name)

    def __getstate__(self) -> Dict[str, Any]:
        """Get a representation of the state of this object.

        This method is invoked by pickle.

        Returns:
            The information to be recorded by a pickle summary of this object.
        """
        state = self.__dict__.copy()
        del state['system_config']
        state['history'] = dict(state['history'])
        return state

    def __setstate__(self, state: Dict[str, Any]) -> None:
        """Set this objects internal state from a dictionary of variables.

        This method is invoked by pickle.

        Args:
            state: The saved state to be used by this object.
        """
        history = defaultdict(lambda: defaultdict(dict))
        history.update(state.get('history', {}))
        state['history'] = history
        self.__dict__.update(state)


def average_summaries(name: str, summaries: List[Summary]) -> Summary:
    """Average multiple summaries together, storing their metric means +- stdevs.

    Args:
        name: The name for the new summary.
        summaries: A list of summaries to be averaged.

    Returns:
        A single summary object reporting mean+-stddev for each metric. If a particular value has only 1 datapoint, it
        will not be averaged.
    """
    if len(summaries) == 0:
        return Summary(name=name)
    if len(summaries) == 1:
        summaries[0].name = name
        return summaries[0]
    consolidated = Summary(name=name)
    # Find all of the modes, keys, and steps over the various summaries
    modes = {mode for summary in summaries for mode in summary.history.keys()}
    keys = {key for summary in summaries for key_pairs in summary.history.values() for key in key_pairs.keys()}
    steps = {
        step
        for summary in summaries for key_pairs in summary.history.values() for val_pair in key_pairs.values()
        for step in val_pair.keys()
    }
    # Average everything
    for mode in modes:
        for key in keys:
            if key == 'epoch':
                # doesn't make sense to average the epoch over different summaries
                # TODO - if all summaries have same epoch dict then preserve it
                continue
            for step in steps:
                vals = []
                for summary in summaries:
                    history = summary.history
                    if mode in history and key in history[mode] and step in history[mode][key]:
                        val = history[mode][key][step]
                        if isinstance(val, str):
                            # Can't plot strings over time...
                            val = [float(s) for s in re.findall(r'(\d+\.\d+|\.?\d+)', val)]
                            if len(val) == 1:
                                # We got an unambiguous number
                                val = val[0]
                            else:
                                val = None
                        elif isinstance(val, ValWithError):
                            val = val.y
                        elif not isinstance(val, (int, float)):
                            val = None
                        if val is not None:
                            vals.append(val)
                if mode == 'test':
                    # We will consolidate these later
                    val = vals
                else:
                    val = _reduce_list(vals)
                if val is None:
                    continue
                consolidated.history[mode][key][step] = val
        if mode == 'test':
            # Due to early stopping, the test mode might be invoked at different steps/epochs. These values will be
            # merged and assigned to the largest available step.
            for key, step_val in consolidated.history[mode].items():
                vals = []
                for step, val in step_val.items():
                    if isinstance(val, list):
                        vals.extend(val)
                    else:
                        vals.append(val)
                step = max(step_val.keys())
                val = _reduce_list(vals)
                consolidated.history[mode][key].clear()
                if val is not None:
                    consolidated.history[mode][key][step] = val
    return consolidated


def _reduce_list(vals: List[Union[int, float]]) -> Union[None, int, float, ValWithError]:
    """Convert a list of numbers into a consolidated summary of those numbers.

    Args:
        vals: A list of values to be summarized.

    Returns:
        None if `vals` is empty, otherwise (mean - std, mean, mean + std) if multiple non-equal values are provided, or
        else simply the value if only a single (possibly repeated) value is found.
    """
    val = None
    if len(vals) > 1:
        if vals[0] == vals[1] and vals.count(vals[0]) == len(vals):
            # If all values are exactly the same, then return single value rather than computing std. The first check is
            # to give fast short-circuiting
            val = vals[0]
        else:
            mean = statistics.mean(vals)
            std = statistics.stdev(vals)
            val = ValWithError(mean - std, mean, mean + std)
    elif len(vals) == 1:
        val = vals[0]
    return val



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.test.unittest_util import OneLayerTorchModel, is_equal, one_layer_tf_model


class TestFeedForward(unittest.TestCase):
    def test_feed_forward_tf(self):
        model = one_layer_tf_model()
        x = tf.constant([[1.0, 1.0, 1.0], [1.0, -1.0, -0.5]])
        obj1 = fe.backend.feed_forward(model, x)
        obj2 = tf.constant([[6.0], [-2.5]])
        self.assertTrue(is_equal(obj1, obj2))

    def test_feed_forward_torch(self):
        model = OneLayerTorchModel()
        x = torch.tensor([[1.0, 1.0, 1.0], [1.0, -1.0, -0.5]])
        obj1 = fe.backend.feed_forward(model, x).detach()
        obj2 = torch.tensor([[6.0], [-2.5]])
        self.assertTrue(is_equal(obj1, obj2))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import torch

from fastestimator.architecture.pytorch import LeNet


class TestLenet(unittest.TestCase):
    def test_lenet_default(self):
        data = np.ones((1, 1, 28, 28))
        input_data = torch.Tensor(data)
        lenet = LeNet()
        output_shape = lenet(input_data).detach().numpy().shape
        self.assertEqual(output_shape, (1, 10))

    def test_lenet_specific_input_shape_classes(self):
        size = (1, 18, 18)
        classes = 3
        data = np.ones((1, ) + size)
        input_data = torch.Tensor(data)
        lenet = LeNet(size, classes=classes)
        output_shape = lenet(input_data).detach().numpy().shape
        self.assertEqual(output_shape, (1, classes))

    def test_check_input_shape(self):
        with self.subTest("length not 3"):
            with self.assertRaises(ValueError):
                LeNet._check_input_shape((1, ))

        with self.subTest("width or height is smaller than 18"):
            with self.assertRaises(ValueError):
                LeNet._check_input_shape((1, 13, 18))

            with self.assertRaises(ValueError):
                LeNet._check_input_shape((1, 18, 1))

        with self.subTest("both are not smaller than 18"):
            LeNet._check_input_shape((1, 18, 18))
            LeNet._check_input_shape((32, 18, 100))
            LeNet._check_input_shape((64, 200, 18))



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np

from fastestimator.summary.logs.log_plot import _MetricGroup


class TestMetricGroups(unittest.TestCase):
    def test_add(self):
        group = _MetricGroup()
        group.add(exp_id=0, mode='train', ds_id='ds1', values={0: 5, 10: 17})
        group.add(exp_id=0, mode='train', ds_id='ds2', values={0: 7, 8: 19})
        group.add(exp_id=1, mode='eval', ds_id='ds2', values={3: 2.6})

        self.assertTrue(np.array_equal(group[0]['train']['ds1'], np.array([[0, 5], [10, 17]])))
        self.assertTrue(np.array_equal(group[0]['train']['ds2'], np.array([[0, 7], [8, 19]])))
        self.assertTrue(np.array_equal(group[1]['eval']['ds2'], np.array([[3, 2.6]])))

    def test_ndim(self):
        with self.subTest("1D"):
            group = _MetricGroup()
            group.add(exp_id=0, mode='train', ds_id='ds1', values={0: 5})
            group.add(exp_id=0, mode='train', ds_id='ds2', values={8: 19})
            group.add(exp_id=1, mode='eval', ds_id='ds2', values={3: 2.6})
            self.assertEqual(group.ndim(), 1)
        with self.subTest("2D"):
            group = _MetricGroup()
            group.add(exp_id=0, mode='train', ds_id='ds1', values={0: 5})
            group.add(exp_id=0, mode='train', ds_id='ds2', values={0: 7, 8: 19})
            group.add(exp_id=1, mode='eval', ds_id='ds2', values={3: 2.6})
            self.assertEqual(group.ndim(), 2)

    def test_modes(self):
        group = _MetricGroup()
        group.add(exp_id=0, mode='train', ds_id='ds1', values={0: 5, 10: 17})
        group.add(exp_id=0, mode='train', ds_id='ds2', values={0: 7, 8: 19})
        group.add(exp_id=1, mode='train', ds_id='ds1', values={0: 5, 10: 17})
        group.add(exp_id=1, mode='eval', ds_id='ds2', values={3: 2.6})

        with self.subTest("1 Mode"):
            self.assertEquals(group.modes(0), ['train'])
        with self.subTest("2 Modes"):
            self.assertEquals(group.modes(1), ['train', 'eval'])

    def test_ds_ids(self):
        group = _MetricGroup()
        group.add(exp_id=0, mode='train', ds_id='ds1', values={0: 5, 10: 17})
        group.add(exp_id=0, mode='train', ds_id='ds2', values={0: 7, 8: 19})
        group.add(exp_id=1, mode='train', ds_id='ds1', values={0: 5, 10: 17})
        group.add(exp_id=1, mode='eval', ds_id='ds3', values={3: 2.6})

        with self.subTest("All Ids Single Mode"):
            self.assertEquals(group.ds_ids(0), ['ds1', 'ds2'])
        with self.subTest("Mode target present"):
            self.assertEquals(group.ds_ids(0, mode='train'), ['ds1', 'ds2'])
        with self.subTest("Mode target absent"):
            self.assertEquals(group.ds_ids(0, mode='test'), [])
        with self.subTest("All Ids Multi Mode"):
            self.assertEquals(group.ds_ids(1), ['ds1', 'ds3'])
        with self.subTest("Mode target present"):
            self.assertEquals(group.ds_ids(1, mode='train'), ['ds1'])
            self.assertEquals(group.ds_ids(1, mode='eval'), ['ds3'])
        with self.subTest("Mode target absent"):
            self.assertEquals(group.ds_ids(1, mode='test'), [])



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple

import torch
import torch.nn as nn
from torch.nn import functional as F
from torch.nn.init import kaiming_normal_ as he_normal


class UNetEncoderBlock(nn.Module):
    """A UNet encoder block.

    This class is intentionally not @traceable (models and layers are handled by a different process).

    Args:
        in_channels: How many channels enter the encoder.
        out_channels: How many channels leave the encoder.
    """
    def __init__(self, in_channels: int, out_channels: int) -> None:
        super().__init__()
        self.layers = nn.Sequential(nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
                                    nn.ReLU(inplace=True),
                                    nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1),
                                    nn.ReLU(inplace=True))

        for layer in self.layers:
            if isinstance(layer, nn.Conv2d):
                he_normal(layer.weight.data)
                layer.bias.data.zero_()

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        out = self.layers(x)
        return out, F.max_pool2d(out, 2)


class UNetDecoderBlock(nn.Module):
    """A UNet decoder block.

    This class is intentionally not @traceable (models and layers are handled by a different process).

    Args:
        in_channels: How many channels enter the decoder.
        mid_channels: How many channels are used for the decoder's intermediate layer.
        out_channels: How many channels leave the decoder.
    """
    def __init__(self, in_channels: int, mid_channels: int, out_channels: int) -> None:
        super().__init__()
        self.layers = nn.Sequential(nn.Conv2d(in_channels, mid_channels, 3, padding=1),
                                    nn.ReLU(inplace=True),
                                    nn.Conv2d(mid_channels, mid_channels, 3, padding=1),
                                    nn.ReLU(inplace=True),
                                    nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True),
                                    nn.Conv2d(mid_channels, out_channels, 3, padding=1),
                                    nn.ReLU(inplace=True))

        for layer in self.layers:
            if isinstance(layer, nn.Conv2d):
                he_normal(layer.weight.data)
                layer.bias.data.zero_()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.layers(x)


class AttentionBlock(nn.Module):
    """An Attention block.

    This class is intentionally not @traceable (models and layers are handled by a different process).

    Args:
        in_channels: How many channels enter the attention block.
        out_channels: How many channels leave the attention block.
    """
    def __init__(self, in_channels: int, out_channels: int) -> None:
        super().__init__()
        self.dec_layers = nn.Sequential(nn.Conv2d(in_channels, out_channels, 1), nn.BatchNorm2d(out_channels))
        self.enc_layers = nn.Sequential(nn.Conv2d(in_channels, out_channels, 1), nn.BatchNorm2d(out_channels))
        self.attn_layers = nn.Sequential(nn.Conv2d(out_channels, 1, 1), nn.BatchNorm2d(1))

    def forward(self, decoder_input: torch.Tensor, encoder_input: torch.Tensor) -> torch.Tensor:
        dec_out = self.dec_layers(decoder_input)
        enc_out = self.enc_layers(encoder_input)
        attn = F.relu(dec_out + enc_out)
        attn = self.attn_layers(attn)
        attn = torch.sigmoid(attn)
        return encoder_input * attn


class AttentionUNet(nn.Module):
    """Attention based UNet implementation in PyTorch.

    This class is intentionally not @traceable (models and layers are handled by a different process).

    Args:
        input_size: The size of the input tensor (channels, height, width).

    Raises:
        ValueError: Length of `input_size` is not 3.
        ValueError: `input_size`[1] or `input_size`[2] is not a multiple of 16.
    """
    def __init__(self, input_size: Tuple[int, int, int] = (1, 128, 128)) -> None:
        AttentionUNet._check_input_size(input_size)
        super().__init__()
        self.input_size = input_size
        self.enc1 = UNetEncoderBlock(in_channels=input_size[0], out_channels=64)
        self.enc2 = UNetEncoderBlock(in_channels=64, out_channels=128)
        self.enc3 = UNetEncoderBlock(in_channels=128, out_channels=256)
        self.enc4 = UNetEncoderBlock(in_channels=256, out_channels=512)
        self.bottle_neck = UNetDecoderBlock(in_channels=512, mid_channels=1024, out_channels=512)
        self.attn4 = AttentionBlock(in_channels=512, out_channels=512)
        self.attn3 = AttentionBlock(in_channels=256, out_channels=256)
        self.attn2 = AttentionBlock(in_channels=128, out_channels=128)
        self.attn1 = AttentionBlock(in_channels=64, out_channels=64)
        self.dec4 = UNetDecoderBlock(in_channels=1024, mid_channels=512, out_channels=256)
        self.dec3 = UNetDecoderBlock(in_channels=512, mid_channels=256, out_channels=128)
        self.dec2 = UNetDecoderBlock(in_channels=256, mid_channels=128, out_channels=64)
        self.dec1 = nn.Sequential(nn.Conv2d(128, 64, 3, padding=1),
                                  nn.ReLU(inplace=True),
                                  nn.Conv2d(64, 64, 3, padding=1),
                                  nn.ReLU(inplace=True),
                                  nn.Conv2d(64, 1, 1),
                                  nn.Sigmoid())

        for layer in self.dec1:
            if isinstance(layer, nn.Conv2d):
                he_normal(layer.weight.data)
                layer.bias.data.zero_()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x1, x_e1 = self.enc1(x)
        x2, x_e2 = self.enc2(x_e1)
        x3, x_e3 = self.enc3(x_e2)
        x4, x_e4 = self.enc4(x_e3)

        x_bottle_neck = self.bottle_neck(x_e4)
        x4_attn = self.attn4(x_bottle_neck, x4)
        x_d4 = self.dec4(torch.cat((x_bottle_neck, x4_attn), 1))
        x3_attn = self.attn3(x_d4, x3)
        x_d3 = self.dec3(torch.cat((x_d4, x3_attn), 1))
        x2_attn = self.attn2(x_d3, x2)
        x_d2 = self.dec2(torch.cat((x_d3, x2_attn), 1))
        x1_attn = self.attn1(x_d2, x1)
        x_out = self.dec1(torch.cat((x_d2, x1_attn), 1))
        return x_out

    @staticmethod
    def _check_input_size(input_size):
        if len(input_size) != 3:
            raise ValueError("Length of `input_size` is not 3 (channel, height, width)")

        _, height, width = input_size

        if height < 16 or not (height / 16.0).is_integer() or width < 16 or not (width / 16.0).is_integer():
            raise ValueError("Both height and width of input_size need to be multiples of 16 (16, 32, 48...)")



#  Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ==============================================================================
from typing import Optional, Iterable, Union

import calibration as cal
import numpy as np

from fastestimator.summary.summary import ValWithError
from fastestimator.trace.meta.per_ds import per_ds
from fastestimator.trace.trace import Trace
from fastestimator.util.data import Data
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import to_number


@per_ds
@traceable()
class CalibrationError(Trace):
    """A trace which computes the calibration error for a given set of predictions.

    Unlike many common calibration error estimation algorithms, this one has actual theoretical bounds on the quality
    of its output: https://arxiv.org/pdf/1909.10155v1.pdf.

    Args:
        true_key: Name of the key that corresponds to ground truth in the batch dictionary.
        pred_key: Name of the key that corresponds to predicted score in the batch dictionary.
        mode: What mode(s) to execute this Trace in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Trace in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        output_name: What to call the output from this trace (for example in the logger output).
        method: Either 'marginal' or 'top-label'. 'marginal' calibration averages the calibration error over each class,
            whereas 'top-label' computes the error based on only the most confident predictions.
        confidence_interval: The calibration error confidence interval to be reported (estimated empirically). Should be
            in the range (0, 100), or else None to omit this extra calculation.
        per_ds: Whether to automatically compute this metric individually for every ds_id it runs on, in addition to
            computing an aggregate across all ds_ids on which it runs. This is automatically False if `output_name`
            contains a "|" character.
    """
    def __init__(self,
                 true_key: str,
                 pred_key: str,
                 mode: Union[None, str, Iterable[str]] = ("eval", "test"),
                 ds_id: Union[None, str, Iterable[str]] = None,
                 output_name: str = "calibration_error",
                 method: str = "marginal",
                 confidence_interval: Optional[int] = None,
                 per_ds: bool = True):
        self.y_true = []
        self.y_pred = []
        assert method in ('marginal', 'top-label'), \
            f"CalibrationError 'method' must be either 'marginal' or 'top-label', but got {method}."
        self.method = method
        if confidence_interval is not None:
            assert 0 < confidence_interval < 100, \
                f"CalibrationError 'confidence_interval' must be between 0 and 100, but got {confidence_interval}."
            confidence_interval = 1.0 - confidence_interval / 100.0
        self.confidence_interval = confidence_interval
        super().__init__(inputs=[true_key, pred_key], outputs=output_name, mode=mode, ds_id=ds_id)
        self.per_ds = per_ds

    @property
    def true_key(self) -> str:
        return self.inputs[0]

    @property
    def pred_key(self) -> str:
        return self.inputs[1]

    def on_epoch_begin(self, data: Data) -> None:
        self.y_true = []
        self.y_pred = []

    def on_batch_end(self, data: Data) -> None:
        y_true, y_pred = to_number(data[self.true_key]), to_number(data[self.pred_key])
        if y_true.shape[-1] > 1 and y_true.ndim > 1:
            y_true = np.argmax(y_true, axis=-1)
        assert y_pred.shape[0] == y_true.shape[0]
        self.y_true.extend(y_true)
        self.y_pred.extend(y_pred)

    def on_epoch_end(self, data: Data) -> None:
        self.y_true = np.squeeze(np.stack(self.y_true))
        self.y_pred = np.stack(self.y_pred)
        mid = round(cal.get_calibration_error(probs=self.y_pred, labels=self.y_true, mode=self.method), 4)
        low = None
        high = None
        if self.confidence_interval is not None:
            low, _, high = cal.get_calibration_error_uncertainties(probs=self.y_pred, labels=self.y_true,
                                                                   mode=self.method,
                                                                   alpha=self.confidence_interval)
            low = round(low, 4)
            high = round(high, 4)
        data.write_with_log(self.outputs[0], ValWithError(low, mid, high) if low is not None else mid)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import inspect
from typing import Any, Dict, List, Union

import numpy as np

from fastestimator.op.numpyop.numpyop import NumpyOp, forward_numpyop
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import to_list


@traceable()
class Fuse(NumpyOp):
    """Run a sequence of NumpyOps as a single Op.

    Args:
        ops: A sequence of NumpyOps to run. They must all share the same mode. It also doesn't support scheduled ops at
            the moment, though the Fuse itself may be scheduled.

    Raises:
        ValueError: If `repeat` or `ops` are invalid.
    """
    def __init__(self, ops: Union[NumpyOp, List[NumpyOp]]) -> None:
        ops = to_list(ops)
        if len(ops) < 1:
            raise ValueError("Fuse requires at least one op")
        inputs = []
        outputs = []
        mode = ops[0].mode
        ds_id = ops[0].ds_id
        for op in ops:
            if op.mode != mode:
                raise ValueError(f"All Fuse ops must share the same mode, but got {mode} and {op.mode}")
            if op.ds_id != ds_id:
                raise ValueError(f"All Fuse ops must share the same ds_id, but got {ds_id} and {op.ds_id}")
            for inp in op.inputs:
                if inp not in inputs and inp not in outputs:
                    inputs.append(inp)
            for out in op.outputs:
                if out not in outputs:
                    outputs.append(out)
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        self.ops = ops

    def __getstate__(self) -> Dict[str, List[Dict[Any, Any]]]:
        return {'ops': [elem.__getstate__() if hasattr(elem, '__getstate__') else {} for elem in self.ops]}

    def set_rua_level(self, magnitude_coef: float) -> None:
        """Set the augmentation intensity based on the magnitude_coef.

        This method is specifically designed to be invoked by the RUA Op.

        Args:
            magnitude_coef: The desired augmentation intensity (range [0-1]).

        Raises:
            AttributeError: If ops don't have a 'set_rua_level' method.
        """
        for op in self.ops:
            if hasattr(op, "set_rua_level") and inspect.ismethod(getattr(op, "set_rua_level")):
                op.set_rua_level(magnitude_coef=magnitude_coef)
            else:
                raise AttributeError(
                    "RUA Augmentations should have a 'set_rua_level' method but it's not present in Op: {}".format(
                        op.__class__.__name__))

    def forward(self, data: List[np.ndarray], state: Dict[str, Any]) -> List[np.ndarray]:
        data = {key: elem for key, elem in zip(self.inputs, data)}
        forward_numpyop(self.ops, data, state)
        return [data[key] for key in self.outputs]



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import tempfile
import unittest

import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.test.unittest_util import MultiLayerTorchModel, is_equal, one_layer_tf_model, sample_system_object
from fastestimator.trace.io import ModelSaver
from fastestimator.util.data import Data


def one_layer_model_without_weights():
    input = tf.keras.layers.Input([3])
    x = tf.keras.layers.Dense(units=1, use_bias=False)(input)
    model = tf.keras.models.Model(inputs=input, outputs=x)
    return model


class MultiLayerTorchModelWithoutWeights(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.fc1 = torch.nn.Linear(4, 2, bias=False)
        self.fc2 = torch.nn.Linear(2, 1, bias=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.fc2(x)
        return x


class TestModelSaver(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.save_dir = tempfile.mkdtemp()

    def test_tf_model(self):
        model = fe.build(model_fn=one_layer_tf_model, optimizer_fn='adam')
        model_saver = ModelSaver(model=model, save_dir=self.save_dir)
        model_saver.system = sample_system_object()
        model_saver.on_epoch_end(data=Data())
        model_name = "{}_epoch_{}".format(model_saver.model.model_name, model_saver.system.epoch_idx)
        tf_model_path = os.path.join(self.save_dir, model_name + '.h5')
        with self.subTest('Check if model is saved'):
            self.assertTrue(os.path.exists(tf_model_path))
        with self.subTest('Validate model weights'):
            m2 = fe.build(model_fn=one_layer_model_without_weights, optimizer_fn='adam')
            fe.backend.load_model(m2, tf_model_path)
            self.assertTrue(is_equal(m2.trainable_variables, model.trainable_variables))

    def test_torch_model(self):
        model = fe.build(model_fn=MultiLayerTorchModel, optimizer_fn='adam')
        model_saver = ModelSaver(model=model, save_dir=self.save_dir)
        model_saver.system = sample_system_object()
        model_name = "{}_epoch_{}".format(model_saver.model.model_name, model_saver.system.epoch_idx)
        torch_model_path = os.path.join(self.save_dir, model_name + '.pt')
        if os.path.exists(torch_model_path):
            os.remove(torch_model_path)
        model_saver.on_epoch_end(data=Data())
        with self.subTest('Check if model is saved'):
            self.assertTrue(os.path.exists(torch_model_path))
        with self.subTest('Validate model weights'):
            m2 = fe.build(model_fn=MultiLayerTorchModelWithoutWeights, optimizer_fn='adam')
            fe.backend.load_model(m2, torch_model_path)
            self.assertTrue(is_equal(list(m2.parameters()), list(model.parameters())))

    def test_max_to_keep_tf(self):
        save_dir = tempfile.mkdtemp()
        model = fe.build(model_fn=one_layer_tf_model, optimizer_fn='adam')
        model_saver = ModelSaver(model=model, save_dir=save_dir, max_to_keep=2)
        model_saver.system = sample_system_object()
        model_saver.on_epoch_end(data=Data())
        model_saver.system.epoch_idx += 1
        model_saver.on_epoch_end(data=Data())
        model_saver.system.epoch_idx += 1
        model_saver.on_epoch_end(data=Data())
        model_name = "{}_epoch_{}".format(model_saver.model.model_name, model_saver.system.epoch_idx)
        tf_model_path1 = os.path.join(save_dir, model_name + '.h5')

        model_saver.system.epoch_idx += 1
        model_saver.on_epoch_end(data=Data())
        model_name = "{}_epoch_{}".format(model_saver.model.model_name, model_saver.system.epoch_idx)
        tf_model_path2 = os.path.join(save_dir, model_name + '.h5')

        with self.subTest('Check only two files are kept'):
            self.assertEqual(len(os.listdir(save_dir)), 2)

        with self.subTest('Check two latest models are kept'):
            self.assertTrue(os.path.exists(tf_model_path1))
            self.assertTrue(os.path.exists(tf_model_path2))

    def test_max_to_keep_tf_architecture(self):
        save_dir = tempfile.mkdtemp()
        model = fe.build(model_fn=one_layer_tf_model, optimizer_fn='adam')
        model_saver = ModelSaver(model=model, save_dir=save_dir, max_to_keep=2, save_architecture=True)
        model_saver.system = sample_system_object()
        model_saver.on_epoch_end(data=Data())
        model_saver.system.epoch_idx += 1
        model_saver.on_epoch_end(data=Data())
        model_saver.system.epoch_idx += 1
        model_saver.on_epoch_end(data=Data())
        model_name = "{}_epoch_{}".format(model_saver.model.model_name, model_saver.system.epoch_idx)
        tf_model_path1 = os.path.join(save_dir, model_name + '.h5')
        tf_architecture_path1 = os.path.join(save_dir, model_name)

        model_saver.system.epoch_idx += 1
        model_saver.on_epoch_end(data=Data())
        model_name = "{}_epoch_{}".format(model_saver.model.model_name, model_saver.system.epoch_idx)
        tf_model_path2 = os.path.join(save_dir, model_name + '.h5')
        tf_architecture_path2 = os.path.join(save_dir, model_name)

        with self.subTest('Check only four files are kept'):
            self.assertEqual(len(os.listdir(save_dir)), 4)

        with self.subTest('Check two latest models are kept'):
            self.assertTrue(os.path.exists(tf_model_path1))
            self.assertTrue(os.path.exists(tf_model_path2))
            self.assertTrue(os.path.exists(tf_architecture_path1))
            self.assertTrue(os.path.isdir(tf_architecture_path1))
            self.assertTrue(os.path.exists(tf_architecture_path2))
            self.assertTrue(os.path.isdir(tf_architecture_path2))

    def test_max_to_keep_torch(self):
        save_dir = tempfile.mkdtemp()
        model = fe.build(model_fn=MultiLayerTorchModel, optimizer_fn='adam')
        model_saver = ModelSaver(model=model, save_dir=save_dir, max_to_keep=2)
        model_saver.system = sample_system_object()
        model_saver.on_epoch_end(data=Data())
        model_saver.system.epoch_idx += 1
        model_saver.on_epoch_end(data=Data())
        model_name = "{}_epoch_{}".format(model_saver.model.model_name, model_saver.system.epoch_idx)
        torch_model_path1 = os.path.join(save_dir, model_name + '.pt')

        model_saver.system.epoch_idx += 1
        model_saver.on_epoch_end(data=Data())
        model_name = "{}_epoch_{}".format(model_saver.model.model_name, model_saver.system.epoch_idx)
        torch_model_path2 = os.path.join(save_dir, model_name + '.pt')

        with self.subTest('Check only two file are kept'):
            self.assertEqual(len(os.listdir(save_dir)), 2)

        with self.subTest('Check two latest model are kept'):
            self.assertTrue(os.path.exists(torch_model_path1))
            self.assertTrue(os.path.exists(torch_model_path2))

    def test_torch_architecture_save(self):
        model = fe.build(model_fn=MultiLayerTorchModel, optimizer_fn='adam')
        save_dir = tempfile.mkdtemp()
        with self.assertRaises(ValueError):
            ModelSaver(model=model, save_dir=save_dir, save_architecture=True)

    def test_max_to_keep_invalid_value(self):
        model = fe.build(model_fn=MultiLayerTorchModel, optimizer_fn='adam')
        save_dir = tempfile.mkdtemp()

        with self.subTest('Check max_to_keep < 0'):
            with self.assertRaises(ValueError):
                ModelSaver(model=model, save_dir=save_dir, max_to_keep=-2)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from copy import deepcopy
from typing import Any, Dict, List, Mapping, Optional, Set

import numpy as np
from torch.utils.data import Dataset

from fastestimator.op.numpyop.numpyop import NumpyOp, forward_numpyop
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import pad_batch


class _DelayedDeepDict(dict):
    """A class to perform delayed deep copying from another dictionary.

    This class is intentionally not @traceable (need quick instantiation).

    This can be useful to reduce memory overhead while protecting the original dictionary entries being overridden. It's
    currently a private class because it doesn't implement the full dictionary spec in an intuitive way.

    Args:
        base: The dictionary to be wrapped.
    """
    warned: Set[str] = set()

    def __init__(self, base: Dict[str, Any]):
        super().__init__()
        self.base = base
        self.mask = set()

    def __getitem__(self, key: str) -> Any:
        if key not in self and key in self.base and key not in self.mask:
            item = self.base[key]
            if isinstance(item, np.ndarray):
                # We'll avoid copying ndarrays for speed/memory savings. The forward_numpyop function will copy later if
                # required.
                item = item.view()
                item.flags.writeable = False
                self[key] = item
            else:
                self[key] = deepcopy(item)
        return super().__getitem__(key)

    def __delitem__(self, key: str):
        # The key will still be in the base dictionary, but that can be handled by the finalize later.
        if key in self or key not in self.base:
            # 'key not in base' to raise errors when key doesn't exist
            super().__delitem__(key)
        self.mask.add(key)  # Mask the key so that it cannot be drawn out of the base

    def finalize(self, retain: Optional[Set[str]] = None, deep_remainder: bool = True) -> None:
        """Finish migrating the data from the original dictionary into this one.

        Args:
            retain: Which keys to keep (or an empty set to keep all available keys).
            deep_remainder: Whether to deep copy any keys which have not yet been copied.
        """
        for key in self.base:
            if key in self.mask:
                continue
            if retain and key not in retain:
                if not key in self.warned:
                    self.warned.add(key)
                    print("FastEstimator-Warn: the key '{}' is being pruned since it is unused outside of the Pipeline."
                          " To prevent this, you can declare the key as an input of a Trace or TensorOp.".format(key))
                continue
            if key not in self:
                if deep_remainder:
                    self[key] = deepcopy(self.base[key])
                else:
                    self[key] = self.base[key]
        if retain:
            for key in self.keys() - retain:
                if not key in self.warned:
                    self.warned.add(key)
                    print("FastEstimator-Warn: the key '{}' is being pruned since it is unused outside of the Pipeline."
                          " To prevent this, you can declare the key as an input of a Trace or TensorOp.".format(key))
                del self[key]
        self.base = {}
        # We need to mark all of the arrays as writeable again to avoid warnings from pytorch. Once torch wraps the
        # arrays, in-place edits on the torch tensors do not impact the numpy arrays anyways.
        for val in self.values():
            if isinstance(val, np.ndarray):
                val.flags.writeable = True


@traceable()
class OpDataset(Dataset):
    """A wrapper for datasets which allows operators to be applied to them in a pipeline.

    This class should not be directly instantiated by the end user. The fe.Pipeline will automatically wrap datasets
    within an Op dataset as needed.

    Args:
        dataset: The base dataset to wrap.
        ops: A list of ops to be applied after the base `dataset` `__getitem__` is invoked.
        mode: What mode the system is currently running in ('train', 'eval', 'test', or 'infer').
        output_keys: What keys can be produced from pipeline. If None, all keys will be considered.
        deep_remainder: Whether data which is not modified by Ops should be deep copied or not. This argument is used to
            help with RAM management, but end users can almost certainly ignore it.
        shuffle: Whether to shuffle batched datasets every epoch.
    """
    def __init__(self,
                 dataset: Dataset,
                 ops: List[NumpyOp],
                 mode: str,
                 output_keys: Optional[Set[str]] = None,
                 deep_remainder: bool = True,
                 shuffle: bool = True) -> None:
        self.dataset = dataset
        if hasattr(self.dataset, "reset_index_maps") and shuffle:
            self.dataset.reset_index_maps()
        self.ops = ops
        self.mode = mode
        self.output_keys = output_keys
        self.deep_remainder = deep_remainder

    def __getitem__(self, index: int) -> Mapping[str, Any]:
        """Fetch a data instance at a specified index, and apply transformations to it.

        Args:
            index: Which datapoint to retrieve.

        Returns:
            The data dictionary from the specified index, with transformations applied.
        """
        item = self.dataset[index]
        if isinstance(item, list):
            # BatchDataset may randomly sample the same elements multiple times, so need to avoid reprocessing
            unique_samples = {}  # id: idx
            results = []
            for idx, data in enumerate(item):
                data_id = id(data)
                if data_id not in unique_samples:
                    data = _DelayedDeepDict(data)
                    forward_numpyop(self.ops, data, {'mode': self.mode})
                    data.finalize(retain=self.output_keys, deep_remainder=self.deep_remainder)
                    results.append(data)
                    unique_samples[data_id] = idx
                else:
                    results.append(results[unique_samples[data_id]])
            if hasattr(self.dataset, "pad_value") and self.dataset.pad_value is not None:
                pad_batch(results, self.dataset.pad_value)
            results = {key: np.array([result[key] for result in results]) for key in results[0]}
        else:
            results = _DelayedDeepDict(item)
            forward_numpyop(self.ops, results, {'mode': self.mode})
            results.finalize(retain=self.output_keys, deep_remainder=self.deep_remainder)
        return results

    def __len__(self):
        return len(self.dataset)



import unittest
from copy import deepcopy

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.test.unittest_util import OneLayerTorchModel, one_layer_tf_model


class TestUpdateModel(unittest.TestCase):
    def test_tf_model_with_get_gradient(self):
        def update(x, model):
            with tf.GradientTape(persistent=True) as tape:
                y = fe.backend.feed_forward(model, x)
                gradient = fe.backend.get_gradient(target=y, sources=model.trainable_variables, tape=tape)
                fe.backend.update_model(model, gradients=gradient)
                return gradient

        lr = 0.1
        model = fe.build(model_fn=one_layer_tf_model, optimizer_fn=lambda: tf.optimizers.SGD(lr))
        init_weights = [x.numpy() for x in model.trainable_variables]
        x = tf.constant([[1, 1, 1], [1, 1, 1]])
        strategy = tf.distribute.get_strategy()
        if isinstance(strategy, tf.distribute.MirroredStrategy):
            gradients = strategy.run(update, args=(x, model))
            gradients = [x.values[0].numpy() for x in gradients]  # from PerReplica to numpy
            # only take the first because other are the same
        else:
            gradients = update(x, model)
            gradients = [x.numpy() for x in gradients]
        gradients_factor = max(torch.cuda.device_count(), 1)  # if multi-gpu, the gradient will accumulate
        new_weights = [x.numpy() for x in model.trainable_variables]
        for init_w, new_w, grad in zip(init_weights, new_weights, gradients):
            new_w_ans = init_w - grad * lr * gradients_factor
            self.assertTrue(np.allclose(new_w_ans, new_w))

    def test_torch_model_with_get_gradient(self):
        lr = 0.1
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        model = fe.build(model_fn=OneLayerTorchModel,
                         optimizer_fn=lambda x: torch.optim.SGD(params=x, lr=lr)).to(device)
        init_weights = [deepcopy(x).cpu().detach().numpy() for x in model.parameters() if x.requires_grad]

        x = torch.tensor([1.0, 1.0, 1.0]).to(torch.device(device))
        y = fe.backend.feed_forward(model.module if torch.cuda.device_count() > 1 else model, x)

        gradients = fe.backend.get_gradient(target=y, sources=[x for x in model.parameters() if x.requires_grad])
        fe.backend.update_model(model, gradients=gradients)

        gradients = [x.cpu().detach().numpy() for x in gradients]
        new_weights = [x.cpu().detach().numpy() for x in model.parameters() if x.requires_grad]

        for init_w, new_w, grad in zip(init_weights, new_weights, gradients):
            new_w_ans = init_w - grad * lr
            self.assertTrue(np.allclose(new_w_ans, new_w))

    def test_tf_model_with_arbitrary_gradient(self):
        def update(gradients, model):
            fe.backend.update_model(model, gradients=gradients)

        lr = 0.1
        model = fe.build(model_fn=one_layer_tf_model, optimizer_fn=lambda: tf.optimizers.SGD(lr))
        init_weights = [x.numpy() for x in model.trainable_variables]
        gradients = [tf.constant([[1.0], [1.0], [1.0]])]

        strategy = tf.distribute.get_strategy()
        if isinstance(strategy, tf.distribute.MirroredStrategy):
            strategy.run(update, args=(gradients, model))
        else:
            update(gradients, model)
        gradients = [x.numpy() for x in gradients]  # tensor to numpy
        new_weights = [x.numpy() for x in model.trainable_variables]
        for init_w, new_w, grad in zip(init_weights, new_weights, gradients):
            new_w_ans = init_w - grad * lr
            self.assertTrue(np.allclose(new_w_ans, new_w))

    def test_torch_model_with_arbitrary_gradient(self):
        lr = 0.1
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        model = fe.build(model_fn=OneLayerTorchModel,
                         optimizer_fn=lambda x: torch.optim.SGD(params=x, lr=lr)).to(device)
        init_weights = [deepcopy(x).cpu().detach().numpy() for x in model.parameters() if x.requires_grad]
        gradients = [torch.tensor([[1.0, 1.0, 1.0]]).to(torch.device(device))]
        fe.backend.update_model(model, gradients=gradients)

        gradients = [x.cpu().detach().numpy() for x in gradients]
        new_weights = [x.cpu().detach().numpy() for x in model.parameters() if x.requires_grad]

        for init_w, new_w, grad in zip(init_weights, new_weights, gradients):
            new_w_ans = init_w - grad * lr
            self.assertTrue(np.allclose(new_w_ans, new_w))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import unittest

import numpy as np

from fastestimator.op.numpyop.univariate import ReadImage
from fastestimator.test.unittest_util import is_equal


class TestReadImage(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.img1_path = os.path.abspath(
            os.path.join(__file__, "..", "..", "..", "..", "util", "resources", "test_read_image.png"))
        cls.img2_path = os.path.abspath(
            os.path.join(__file__, "..", "..", "..", "..", "util", "resources", "test_read_image2.png"))
        cls.expected_image_output = np.zeros((28, 28, 3))
        cls.expected_second_image_output = 255 * np.ones((28, 28, 3))

    def test_single_input(self):
        data = [self.img1_path]
        image = ReadImage(inputs='x', outputs='x')
        output = image.forward(data=data, state={})
        self.assertTrue(is_equal(output[0], self.expected_image_output))

    def test_multi_input(self):
        data = [self.img1_path, self.img2_path]
        image = ReadImage(inputs='x', outputs='x')
        output = image.forward(data=data, state={})
        with self.subTest('Check first image in data'):
            self.assertTrue(is_equal(output[0], self.expected_image_output))
        with self.subTest('Check second image in data'):
            self.assertTrue(is_equal(output[1], self.expected_second_image_output))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.test.unittest_util import is_equal


class TestPermute(unittest.TestCase):
    def test_permute_np(self):
        n = np.array([[[0, 1], [2, 3]], [[4, 5], [6, 7]], [[8, 9], [10, 11]]])
        with self.subTest("permutation case 1"):
            obj1 = fe.backend.permute(n, [2, 0, 1])
            obj2 = np.array([[[0, 2], [4, 6], [8, 10]], [[1, 3], [5, 7], [9, 11]]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("permutation case 2"):
            obj1 = fe.backend.permute(n, [0, 2, 1])
            obj2 = np.array([[[0, 2], [1, 3]], [[4, 6], [5, 7]], [[8, 10], [9, 11]]])
            self.assertTrue(is_equal(obj1, obj2))

    def test_permute_tf(self):
        t = tf.constant([[[0, 1], [2, 3]], [[4, 5], [6, 7]], [[8, 9], [10, 11]]])
        with self.subTest("permutation case 1"):
            obj1 = fe.backend.permute(t, [2, 0, 1])
            obj2 = tf.constant([[[0, 2], [4, 6], [8, 10]], [[1, 3], [5, 7], [9, 11]]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("permutation case 2"):
            obj1 = fe.backend.permute(t, [0, 2, 1])
            obj2 = tf.constant([[[0, 2], [1, 3]], [[4, 6], [5, 7]], [[8, 10], [9, 11]]])
            self.assertTrue(is_equal(obj1, obj2))

    def test_permute_torch(self):
        t = torch.tensor([[[0, 1], [2, 3]], [[4, 5], [6, 7]], [[8, 9], [10, 11]]])
        with self.subTest("permutation case 1"):
            obj1 = fe.backend.permute(t, [2, 0, 1])
            obj2 = torch.tensor([[[0, 2], [4, 6], [8, 10]], [[1, 3], [5, 7], [9, 11]]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("permutation case 2"):
            obj1 = fe.backend.permute(t, [0, 2, 1])
            obj2 = torch.tensor([[[0, 2], [1, 3]], [[4, 6], [5, 7]], [[8, 10], [9, 11]]])
            self.assertTrue(is_equal(obj1, obj2))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import math
from typing import List, TypeVar, Union

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
import torch

from fastestimator.util.util import to_list

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


def percentile(tensor: Tensor,
               percentiles: Union[int, List[int]],
               axis: Union[None, int, List[int]] = None,
               keepdims: bool = True) -> Tensor:
    """Compute the `percentiles` of a `tensor`.

    The n-th percentile of `tensor` is the value n/100 of the way from the minimum to the maximum in a sorted copy of
    `tensor`. If the percentile falls in between two values, the lower of the two values will be used.

    This method can be used with Numpy data:
    ```python
    n = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    b = fe.backend.percentile(n, percentiles=[66])  # [[[6]]]
    b = fe.backend.percentile(n, percentiles=[66], axis=0)  # [[[4, 5, 6]]]
    b = fe.backend.percentile(n, percentiles=[66], axis=1)  # [[[2], [5], [8]]]
    ```

    This method can be used with TensorFlow tensors:
    ```python
    t = tf.constant([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    b = fe.backend.percentile(t, percentiles=[66])  # [[[6]]]
    b = fe.backend.percentile(t, percentiles=[66], axis=0)  # [[[4, 5, 6]]]
    b = fe.backend.percentile(t, percentiles=[66], axis=1)  # [[[2], [5], [8]]]
    ```

    This method can be used with PyTorch tensors:
    ```python
    p = tf.constant([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    b = fe.backend.percentile(p, percentiles=[66])  # [[[6]]]
    b = fe.backend.percentile(p, percentiles=[66], axis=0)  # [[[4, 5, 6]]]
    b = fe.backend.percentile(p, percentiles=[66], axis=1)  # [[[2], [5], [8]]]
    ```

    Args:
        tensor: The tensor from which to extract percentiles.
        percentiles: One or more percentile values to be computed.
        axis: Along which axes to compute the percentile (None to compute over all axes).
        keepdims: Whether to maintain the number of dimensions from `tensor`.

    Returns:
        The `percentiles` of the given `tensor`.

    Raises:
        ValueError: If `tensor` is an unacceptable data type.
    """
    if tf.is_tensor(tensor):
        if isinstance(percentiles, List):
            percentiles = tf.convert_to_tensor(percentiles)
        return tfp.stats.percentile(tensor, percentiles, axis=axis, keep_dims=keepdims, interpolation='lower')
    elif isinstance(tensor, torch.Tensor):
        n_dims = len(tensor.shape)
        if axis is None:
            # Default behavior in tf without axis is to compress all dimensions
            axis = list(range(n_dims))
        # Convert negative axis values to their positive counterparts
        if isinstance(axis, int):
            axis = [axis]
        for idx, elem in enumerate(axis):
            axis[idx] = elem % n_dims
        # Extract dims which are not being considered
        other_dims = sorted(set(range(n_dims)).difference(axis))
        # Flatten all of the permutation axis down for kth-value computation
        permutation = other_dims + list(axis)
        permuted = tensor.permute(*permutation)
        other_shape = [tensor.shape[i] for i in other_dims]
        other_shape.append(np.prod([tensor.shape[i] for i in axis]))
        permuted = torch.reshape(permuted, other_shape)
        results = []
        for tile in to_list(percentiles):
            target = 1 + math.floor(tile / 100.0 * (permuted.shape[-1] - 1))
            kth_val = torch.kthvalue(permuted, k=target, dim=-1, keepdim=True)[0]
            for dim in range(n_dims - len(kth_val.shape)):
                kth_val = torch.unsqueeze(kth_val, dim=-1)
            # Undo the permutation from earlier
            kth_val = kth_val.permute(*np.argsort(permutation))
            if not keepdims:
                for dim in reversed(axis):
                    kth_val = torch.squeeze(kth_val, dim=dim)
            results.append(kth_val)
        if isinstance(percentiles, int):
            return results[0]
        else:
            return torch.stack(results, dim=0)
    elif isinstance(tensor, np.ndarray):
        return np.percentile(tensor, percentiles, axis=axis, keepdims=keepdims, interpolation='lower')
    else:
        raise ValueError("Unrecognized tensor type {}".format(type(tensor)))



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from torch.utils.data import DataLoader
from torch.utils.data.dataloader import _BaseDataLoaderIter, _MultiProcessingDataLoaderIter


class FEDataLoader(DataLoader):
    _current_threads = []

    def shutdown(self) -> None:
        """Close the worker threads used by this iterator.

        The hope is that this will prevent "RuntimeError: DataLoader worker (pid(s) XXXX) exited unexpectedly" during
        the test suites.
        """
        if isinstance(self._iterator, _MultiProcessingDataLoaderIter):
            self._iterator._shutdown_workers()
        self._iterator = None
        FEDataLoader._current_threads.clear()

    def __iter__(self) -> _BaseDataLoaderIter:
        # Similar to the original iter method, but we remember iterators in order to manually close them when new ones
        # are created
        if self.persistent_workers and self.num_workers > 0:
            if self._iterator is None:
                self._iterator = self._get_iterator()
            else:
                self._iterator._reset(self)
        else:
            self.shutdown()
            self._iterator = self._get_iterator()
        if isinstance(self._iterator, _MultiProcessingDataLoaderIter):
            FEDataLoader._current_threads.extend([w.pid for w in self._iterator._workers])
        return self._iterator



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import math
import unittest
from io import StringIO
from unittest.mock import patch

import fastestimator as fe
from fastestimator.test.unittest_util import MultiLayerTorchModel, one_layer_tf_model, sample_system_object
from fastestimator.trace.adapt import ReduceLROnPlateau
from fastestimator.util.data import Data


class TestReduceLROnPlateau(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data = Data({'loss': 10})
        cls.tf_expected_msg = "FastEstimator-ReduceLROnPlateau: learning rate reduced to 0.00010000000474974513"
        cls.torch_expected_msg = "FastEstimator-ReduceLROnPlateau: learning rate reduced to 9.999999747378752e-05"

    def test_tf_model_on_epoch_end_reduce_lr(self):
        model = fe.build(model_fn=one_layer_tf_model, optimizer_fn='adam')
        model_name = model.model_name + '_lr'
        lr_on_plateau = ReduceLROnPlateau(model=model, metric='loss')
        lr_on_plateau.system = sample_system_object()
        lr_on_plateau.best = 5
        lr_on_plateau.wait = 11
        with patch('sys.stdout', new=StringIO()) as fake_stdout:
            lr_on_plateau.on_epoch_end(data=self.data)
            log = fake_stdout.getvalue().strip()
            self.assertEqual(log, self.tf_expected_msg)
        with self.subTest('Check learning rate in data'):
            self.assertTrue(math.isclose(self.data[model_name], 0.000100000005, rel_tol=1e-3))

    def test_tf_model_on_epoch_end_lr_wait(self):
        model = fe.build(model_fn=one_layer_tf_model, optimizer_fn='adam')
        lr_on_plateau = ReduceLROnPlateau(model=model, metric='loss')
        lr_on_plateau.system = sample_system_object()
        lr_on_plateau.best = 12
        lr_on_plateau.on_epoch_end(data=self.data)
        with self.subTest('Check value of wait'):
            self.assertEqual(lr_on_plateau.wait, 0)
        with self.subTest('Check value of best'):
            self.assertEqual(lr_on_plateau.best, 10)

    def test_torch_model_on_epoch_end_reduce_lr(self):
        model = fe.build(model_fn=MultiLayerTorchModel, optimizer_fn='adam')
        model_name = model.model_name + '_lr'
        lr_on_plateau = ReduceLROnPlateau(model=model, metric='loss')
        lr_on_plateau.system = sample_system_object()
        lr_on_plateau.best = 5
        lr_on_plateau.wait = 11
        with patch('sys.stdout', new=StringIO()) as fake_stdout:
            lr_on_plateau.on_epoch_end(data=self.data)
            log = fake_stdout.getvalue().strip()
            self.assertEqual(log, self.torch_expected_msg)
        with self.subTest('Check learning rate in data'):
            self.assertTrue(math.isclose(self.data[model_name], 0.000100000005, rel_tol=1e-3))

    def test_torch_model_on_epoch_end_lr_wait(self):
        model = fe.build(model_fn=MultiLayerTorchModel, optimizer_fn='adam')
        lr_on_plateau = ReduceLROnPlateau(model=model, metric='loss')
        lr_on_plateau.system = sample_system_object()
        lr_on_plateau.best = 12
        lr_on_plateau.on_epoch_end(data=self.data)
        with self.subTest('Check value of wait'):
            self.assertEqual(lr_on_plateau.wait, 0)
        with self.subTest('Check value of best'):
            self.assertEqual(lr_on_plateau.best, 10)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict, Iterable, List, Optional, TypeVar, Union

import tensorflow as tf
import tensorflow_probability as tfp
import torch

from fastestimator.backend.maximum import maximum
from fastestimator.backend.reshape import reshape
from fastestimator.backend.roll import roll
from fastestimator.op.tensorop.tensorop import TensorOp

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)


class MixUpBatch(TensorOp):
    """MixUp augmentation for tensors.

    This class should be used in conjunction with MixLoss to perform mix-up training, which helps to reduce
    over-fitting, stabilize GAN training, and against adversarial attacks (https://arxiv.org/abs/1710.09412).

    Args:
        inputs: Key of the input to be mixed up.
        outputs: Key to store the mixed-up outputs.
        mode: What mode to execute in. Probably 'train'.
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        alpha: The alpha value defining the beta distribution to be drawn from during training.
        shared_beta: Sample a single beta for a batch or element wise beta for each image.

    Raises:
        AssertionError: If input arguments are invalid.
    """
    def __init__(self,
                 inputs: Union[str, Iterable[str]],
                 outputs: Iterable[str],
                 mode: Union[None, str, Iterable[str]] = 'train',
                 ds_id: Union[None, str, Iterable[str]] = None,
                 alpha: float = 1.0,
                 shared_beta: bool = True):
        assert alpha > 0, "MixUp alpha value must be greater than zero"
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        assert len(self.outputs) == len(self.inputs) + 1, "MixUpBatch requires 1 more output than inputs"
        self.alpha = alpha
        self.beta = None
        self.shared_beta = shared_beta
        self.in_list, self.out_list = True, True

    def build(self, framework: str, device: Optional[torch.device] = None) -> None:
        if framework == 'tf':
            self.beta = tfp.distributions.Beta(self.alpha, self.alpha)
        elif framework == 'torch':
            self.beta = torch.distributions.beta.Beta(self.alpha, self.alpha)
        else:
            raise ValueError("unrecognized framework: {}".format(framework))

    def forward(self, data: List[Tensor], state: Dict[str, Any]) -> List[Tensor]:
        if self.shared_beta:
            lam = self.beta.sample()
        else:
            lam = self.beta.sample(sample_shape=(data[0].shape[0], ))
            shape = [-1] + [1] * (len(data[0].shape) - 1)
            lam = reshape(lam, shape)
        lam = maximum(lam, (1 - lam))
        mix = [lam * elem + (1.0 - lam) * roll(elem, shift=1, axis=0) for elem in data]
        return mix + [lam]



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

from fastestimator.backend import roll
from fastestimator.test.unittest_util import is_equal


class TestRoll(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_np = np.array([[1.0, 2.0, 3.0], [5.0, 6.0, 7.0]])
        cls.test_output_np_1d = np.array([[5, 6, 7], [1, 2, 3]])
        cls.test_output_np_2d = np.array([[6, 7, 5], [2, 3, 1]])
        cls.test_tf = tf.constant([[1.0, 2.0, 3.0], [5.0, 6.0, 7.0]])
        cls.test_output_tf_1d = tf.constant([[5, 6, 7], [1, 2, 3]])
        cls.test_output_tf_2d = tf.constant([[6, 7, 5], [2, 3, 1]])
        cls.test_torch = torch.Tensor([[1.0, 2.0, 3.0], [5.0, 6.0, 7.0]])
        cls.test_output_torch_1d = torch.Tensor([[5, 6, 7], [1, 2, 3]])
        cls.test_output_torch_2d = torch.Tensor([[6, 7, 5], [2, 3, 1]])

    def test_roll_np_type(self):
        output = roll(self.test_np, shift=1, axis=0)
        self.assertIsInstance(output, np.ndarray, 'Output type must be NumPy array')

    def test_roll_np_1d_shift(self):
        output = roll(self.test_np, shift=1, axis=0)
        self.assertTrue(np.array_equal(output, self.test_output_np_1d))

    def test_roll_np_2d_shift(self):
        output = roll(self.test_np, shift=[-1, -1], axis=[0, 1])
        self.assertTrue(np.array_equal(output, self.test_output_np_2d))

    def test_roll_tf_type(self):
        output = roll(self.test_tf, shift=0, axis=1)
        self.assertIsInstance(output, tf.Tensor, 'Output type must be tf.Tensor')

    def test_roll_tf_1d_shift(self):
        output = roll(self.test_tf, shift=1, axis=0)
        self.assertTrue(is_equal(output, self.test_output_tf_1d))

    def test_roll_tf_2d_shift(self):
        output = roll(self.test_tf, shift=[-1, -1], axis=[0, 1])
        self.assertTrue(is_equal(output, self.test_output_tf_2d))

    def test_sign_torch_type(self):
        output = roll(self.test_torch, shift=1, axis=0)
        self.assertIsInstance(output, torch.Tensor, 'Output type must be torch.Tensor')

    def test_roll_torch_1d_shift(self):
        output = roll(self.test_torch, shift=1, axis=0)
        self.assertTrue(is_equal(output, self.test_output_torch_1d))

    def test_roll_torch_2d_shift(self):
        output = roll(self.test_torch, shift=[-1, -1], axis=[0, 1])
        self.assertTrue(is_equal(output, self.test_output_torch_2d))



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple

import tensorflow as tf
from tensorflow.python.keras import layers
from tensorflow.python.keras.engine.keras_tensor import KerasTensor


def ResNet9(input_size: Tuple[int, int, int] = (32, 32, 3), classes: int = 10) -> tf.keras.Model:
    """A small 9-layer ResNet Tensorflow model for cifar10 image classification.
    The model architecture is from https://github.com/davidcpage/cifar10-fast

    Args:
        input_size: The size of the input tensor (height, width, channels).
        classes: The number of outputs the model should generate.

    Raises:
        ValueError: Length of `input_size` is not 3.
        ValueError: `input_size`[0] or `input_size`[1] is not a multiple of 16.

    Returns:
        A TensorFlow ResNet9 model.
    """
    _check_input_size(input_size)

    # prep layers
    inp = layers.Input(shape=input_size)
    x = layers.Conv2D(64, 3, padding='same')(inp)
    x = layers.BatchNormalization(momentum=0.8)(x)
    x = layers.LeakyReLU(alpha=0.1)(x)
    # layer1
    x = layers.Conv2D(128, 3, padding='same')(x)
    x = layers.MaxPool2D()(x)
    x = layers.BatchNormalization(momentum=0.8)(x)
    x = layers.LeakyReLU(alpha=0.1)(x)
    x = layers.Add()([x, residual(x, 128)])
    # layer2
    x = layers.Conv2D(256, 3, padding='same')(x)
    x = layers.MaxPool2D()(x)
    x = layers.BatchNormalization(momentum=0.8)(x)
    x = layers.LeakyReLU(alpha=0.1)(x)
    # layer3
    x = layers.Conv2D(512, 3, padding='same')(x)
    x = layers.MaxPool2D()(x)
    x = layers.BatchNormalization(momentum=0.8)(x)
    x = layers.LeakyReLU(alpha=0.1)(x)
    x = layers.Add()([x, residual(x, 512)])
    # layers4
    x = layers.GlobalMaxPool2D()(x)
    x = layers.Flatten()(x)
    x = layers.Dense(classes)(x)
    x = layers.Activation('softmax', dtype='float32')(x)
    model = tf.keras.Model(inputs=inp, outputs=x)

    return model


def residual(x: KerasTensor, num_channel: int) -> KerasTensor:
    """A ResNet unit for ResNet9.

    Args:
        x: Input Keras tensor.
        num_channel: The number of layer channel.

    Return:
        Output Keras tensor.
    """
    x = layers.Conv2D(num_channel, 3, padding='same')(x)
    x = layers.BatchNormalization(momentum=0.8)(x)
    x = layers.LeakyReLU(alpha=0.1)(x)
    x = layers.Conv2D(num_channel, 3, padding='same')(x)
    x = layers.BatchNormalization(momentum=0.8)(x)
    x = layers.LeakyReLU(alpha=0.1)(x)
    return x


def _check_input_size(input_size):
    if len(input_size) != 3:
        raise ValueError("Length of `input_size` is not 3 (channel, height, width)")

    height, width, _ = input_size

    if height < 16 or width < 16:
        raise ValueError("Both height and width of input_size need to not smaller than 16")



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np

from fastestimator.op.numpyop.univariate import InvertImg


class TestInvertImg(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.single_input = [np.random.rand(28, 28, 3)]
        cls.single_output_shape = (28, 28, 3)
        cls.multi_input = [np.random.rand(28, 28, 3), np.random.rand(28, 28, 3)]
        cls.multi_output_shape = (28, 28, 3)

    def test_single_input(self):
        invert_img = InvertImg(inputs='x', outputs='x')
        output = invert_img.forward(data=self.single_input, state={})
        with self.subTest('Check output type'):
            self.assertEqual(type(output), list)
        with self.subTest('Check output image shape'):
            self.assertEqual(output[0].shape, self.single_output_shape)

    def test_multi_input(self):
        invert_img = InvertImg(inputs='x', outputs='x')
        output = invert_img.forward(data=self.multi_input, state={})
        with self.subTest('Check output type'):
            self.assertEqual(type(output), list)
        with self.subTest('Check output list length'):
            self.assertEqual(len(output), 2)
        for img_output in output:
            with self.subTest('Check output mask shape'):
                self.assertEqual(img_output.shape, self.multi_output_shape)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

from fastestimator.op.tensorop.loss import Hinge


class TestHinge(unittest.TestCase):
    def test_tf(self):
        true = tf.constant([[-1, 1, 1, -1], [1, 1, 1, 1], [-1, -1, 1, -1], [1, -1, -1, -1]])
        pred = tf.constant([[0.1, 0.9, 0.05, 0.05], [0.1, -0.2, 0.0, -0.7], [0.0, 0.15, 0.8, 0.05],
                            [1.0, -1.0, -1.0, -1.0]])
        hinge = Hinge(inputs=('x1', 'x2'), outputs='x')
        hinge.build('tf', None)
        output = hinge.forward(data=[pred, true], state={})
        self.assertTrue(np.allclose(output.numpy(), 0.7125))

    def test_torch(self):
        true = torch.tensor([[-1, 1, 1, -1], [1, 1, 1, 1], [-1, -1, 1, -1], [1, -1, -1, -1]])
        pred = torch.tensor([[0.1, 0.9, 0.05, 0.05], [0.1, -0.2, 0.0, -0.7], [0.0, 0.15, 0.8, 0.05],
                             [1.0, -1.0, -1.0, -1.0]])
        hinge = Hinge(inputs=('x1', 'x2'), outputs='x')
        hinge.build('torch', torch.device('cpu'))
        output = hinge.forward(data=[pred, true], state={})
        self.assertTrue(np.allclose(output.numpy(), 0.7125))



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

from fastestimator.search.grid_search import GridSearch


class TestSearch(unittest.TestCase):
    def test_non_dict_params(self):
        with self.assertRaises(AssertionError):
            GridSearch(score_fn=lambda search_idx, x: x, params=False)

    def test_dict_params_wrong_args(self):
        with self.assertRaises(AssertionError):
            GridSearch(score_fn=lambda search_idx, x: x, params={"lr": [1, 2, 3]})

    def test_correct_output(self):
        search = GridSearch(score_fn=lambda search_idx, a, b: a + b, params={"a": [1, 2, 3], "b": [4, 5, 6]})
        search.fit()
        self.assertEqual(search.get_best_results(), ({"a": 3, "b": 6, "search_idx": 9}, 9))

    def test_notebook_restart(self):
        search = GridSearch(score_fn=lambda search_idx, a, b: a + b, params={"a": [1, 2, 3], "b": [4, 5, 6]})
        search.fit()
        search.fit()
        self.assertEqual(search.search_idx, 9)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

from fastestimator.backend import squeeze


class TestSqueeze(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_np = np.array([[[[1], [2]]], [[[3], [3]]]])
        cls.test_output_np = np.array([[1, 2], [3, 3]])
        cls.test_output_np_axis = np.array([[[1], [2]], [[3], [3]]])
        cls.test_tf = tf.constant([[[1], [1]], [[2], [-3]], [[4], [1]]])
        cls.test_output_tf = tf.constant([[1, 1], [2, -3], [4, 1]])
        cls.test_torch = torch.Tensor([[[1], [1]], [[2], [-3]], [[4], [1]]])
        cls.test_output_torch = torch.Tensor([[1, 1], [2, -3], [4, 1]])

    def test_squeeze_np_type(self):
        self.assertIsInstance(squeeze(self.test_np), np.ndarray, 'Output type must be NumPy array')

    def test_squeeze_np_value(self):
        self.assertTrue(np.array_equal(squeeze(self.test_np), self.test_output_np))

    def test_squeeze_tf_type(self):
        self.assertIsInstance(squeeze(self.test_tf), tf.Tensor, 'Output type must be tf.Tensor')

    def test_squeeze_tf_value(self):
        self.assertTrue(np.array_equal(squeeze(self.test_tf).numpy(), self.test_output_tf))

    def test_squeeze_torch_type(self):
        self.assertIsInstance(squeeze(self.test_torch), torch.Tensor, 'Output type must be torch.Tensor')

    def test_squeeze_torch_value(self):
        self.assertTrue(np.array_equal(squeeze(self.test_torch).numpy(), self.test_output_torch))

    def test_squeeze_axis(self):
        self.assertTrue(np.array_equal(squeeze(self.test_np, axis=1), self.test_output_np_axis))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Iterable, Optional, Union

from pylatex import NoEscape, Package, escape_latex
from pylatex.base_classes import Container, ContainerCommand, Environment, LatexObject, Options
from pylatex.lists import Enumerate
from pylatex.utils import bold, dumps_list

from fastestimator.util.util import FEID


class ContainerList(Container):
    """A class to expedite combining pieces of latex together.

    This class is intentionally not @traceable.
    """
    def dumps(self) -> str:
        """Get a string representation of this container.

        Returns:
            A string representation of itself.
        """
        return self.dumps_content()


class PyContainer(ContainerList):
    """A class to convert python containers to a LaTeX representation.

    This class is intentionally not @traceable.

    Args:
        data: The python object to be converted to LaTeX.
        truncate: How many values to display before truncating with an ellipsis. This should be a positive integer or
            None to disable truncation.
    """
    def __init__(self, data: Union[list, tuple, set, dict], truncate: Optional[int] = None):
        self.packages.add(Package('enumitem', options='inline'))
        assert isinstance(data, (list, tuple, set, dict)), f"Unacceptable data type for PyContainer: {type(data)}"
        open_char = '[' if isinstance(data, list) else '(' if isinstance(data, tuple) else r'\{'
        close_char = ']' if isinstance(data, list) else ')' if isinstance(data, tuple) else r'\}'
        ltx = Enumerate(options=Options(NoEscape('label={}'), NoEscape('itemjoin={,}')))
        ltx._star_latex_name = True  # Converts this to an inline list
        self.raw_input = data
        if isinstance(data, dict):
            for key, val in list(data.items())[:truncate]:
                ltx.add_item(ContainerList(data=[key, ": ", val]))
        else:
            for val in list(data)[:truncate]:
                ltx.add_item(val)
        if truncate and len(data) > truncate:
            ltx.add_item(NoEscape(r'\ldots'))
        super().__init__(data=[NoEscape(open_char), ltx, NoEscape(close_char)])


class Verbatim(Environment):
    """A class to put a string inside the latex verbatim environment.

    This class is intentionally not @traceable.

    Args:
        data: The string to be wrapped.
    """
    def __init__(self, data: str):
        super().__init__(options=None, arguments=None, start_arguments=None, data=NoEscape(data))
        self.content_separator = '\n'


class Center(Environment):
    """A class to center content in a page.

    This class is intentionally not @traceable.
    """
    pass


class AdjustBox(Environment):
    """A class to adjust the size of boxes.

    This class is intentionally not @traceable.
    """
    packages = [Package('adjustbox')]


class Form(Environment):
    """A class to allow Form elements.

    This class is intentionally not @traceable. Only one Form is allowed per document.
    """
    _latex_name = 'Form'
    packages = [Package('hyperref', options='hidelinks')]


class TextField(ContainerCommand):
    """A class to create editable text fields.

    This class is intentionally not @traceable. It can only be used inside of a Form.
    """
    _latex_name = "TextField"


class TextFieldBox(ContainerList):
    """A class to wrap TextFields into padded boxes for use in nesting within tables.

    Args:
        name: The name to assign to this TextField. It should be unique within the document since changes to one box
            will impact all boxes with the same name.
        height: How tall should the TextField box be? Note that it will be wrapped by 10pt space on the top and bottom.
    """
    packages = [Package('xcolor', options='table')]

    def __init__(self, name: str, height: str = '2.5cm'):
        data = [
            NoEscape(r"\begin{minipage}{\linewidth}"),
            NoEscape(r"\vspace{3pt}"),
            TextField(options=[
                NoEscape(r'width=\linewidth'),
                NoEscape(f'height={height}'),
                NoEscape('backgroundcolor={0.97 0.97 0.97}'),
                'bordercolor=white',
                'multiline=true',
                f'name={name}'
            ]),
            NoEscape(r"\vspace{3pt}"),
            NoEscape(r"\end{minipage}")
        ]
        super().__init__(data=data)


class HrefFEID(ContainerList):
    """A class to represent a colored and underlined hyperref based on a given fe_id.

    This class is intentionally not @traceable.

    Args:
        fe_id: The id used to link this hyperref.
        name: A string suffix to be printed as part of the link text.
        link_prefix: The prefix for the hyperlink.
        id_in_name: Whether to include the id in front of the name text.
        bold_name: Whether to bold the name.
    """
    def __init__(self,
                 fe_id: FEID,
                 name: str,
                 link_prefix: str = 'tbl',
                 id_in_name: bool = True,
                 bold_name: bool = False):
        self.content_separator = ''
        self.packages.add(Package('hyperref', options='hidelinks'))
        self.packages.add(Package('ulem'))
        self.packages.add(Package('xcolor', options='table'))
        self.fe_id = fe_id
        self.name = name
        data = [
            NoEscape(r'\hyperref['), escape_latex(f"{link_prefix}:"), fe_id, NoEscape(r']{\textcolor{blue}{\uline{')
        ]
        if id_in_name:
            data.append(fe_id)
            if name:
                data.append(": ")
        if name:
            data.append(bold(escape_latex(name)) if bold_name else escape_latex(name))
        data.append(NoEscape("}}}"))
        super().__init__(data=data)


class IterJoin(Container):
    """A class to convert an iterable to a latex representation.

    Args:
        data: Data of the cell.
        token: String to serve as separator among items of `data`.
    """
    def __init__(self, data: Iterable, token: str):
        super().__init__(data=data)
        self.token = token

    def dumps(self) -> str:
        """Get a string representation of this cell.

        Returns:
            A string representation of itself.
        """
        return dumps_list(self, token=self.token)


class WrapText(LatexObject):
    """A class to convert strings or numbers to wrappable latex representation.

    This class will first convert the data to string, and then to a wrappable latex representation if its length is too
    long. This fixes an issue which prevents the first element placed into a latex X column from wrapping correctly.

    Args:
        data: Input data to be converted.
        threshold: When the length of `data` is greater than `threshold`, the resulting string will be made wrappable.

    Raises:
        AssertionError: If `data` is not a string, int, or float.
    """
    def __init__(self, data: Union[str, int, float], threshold: int):
        assert isinstance(data, (str, int, float)), "the self.data type needs to be str, int, float"
        self.threshold = threshold
        self.data = str(data)
        super().__init__()

    def dumps(self) -> str:
        """Get a string representation of this cell.

        Returns:
            A string representation of itself.
        """
        if len(self.data) > self.threshold:
            return NoEscape(r'\seqsplit{' + escape_latex(self.data) + '}')
        else:
            return escape_latex(self.data)



#  Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ==============================================================================
import tempfile
import unittest

import tensorflow as tf

from fastestimator.test.unittest_util import sample_system_object
from fastestimator.trace.xai import LabelTracker
from fastestimator.util import Data


class TestLabelTracker(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Epoch 1 Data
        batch1_metrics = tf.constant([0.9, 0.8, 0.1, 0.4, 0.9, 0.6, 0.1, 0.6, 0.8, 0.3])
        batch1_labels = tf.constant([2, 1, 0, 1, 0, 1, 1, 1, 0, 0])
        batch2_metrics = tf.constant([0.3, 0.1, 0.0, 0.8, 0.5, 0.5, 0.6, 0.2, 0.5, 1.0])
        batch2_labels = tf.constant([1, 1, 1, 1, 0, 1, 0, 0, 0, 0])
        batch3_metrics = tf.constant([0.3, 0.4, 0.7, 0.9, 0.3, 0.0, 0.3, 0.7, 0.8, 0.6])
        batch3_labels = tf.constant([2, 2, 2, 2, 2, 0, 2, 2, 2, 2])
        # Epoch 2 Data
        batch4_metrics = tf.constant([0.8, 0.9, 0.6, 0.1, 0.7, 0.3, 0.9, 0.9, 0.4, 0.6])
        batch4_labels = tf.constant([0, 1, 0, 1, 0, 1, 1, 1, 0, 0])
        batch5_metrics = tf.constant([0.4, 0.4, 0.9, 0.1, 0.4, 0.9, 0.0, 0.8, 1.0, 0.1])
        batch5_labels = tf.constant([2, 2, 2, 2, 0, 1, 0, 0, 0, 0])
        batch6_metrics = tf.constant([0.3, 0.9, 0.9, 0.5, 0.9, 0.8, 0.6, 0.1, 0.1, 0.2])
        batch6_labels = tf.constant([1, 1, 1, 1, 2, 2, 2, 2, 2, 2])

        cls.training_data = [[
            Data({'acc': batch1_metrics, 'y': batch1_labels}),
            Data({'acc': batch2_metrics, 'y': batch2_labels}),
            Data({'acc': batch3_metrics, 'y': batch3_labels})
        ], [Data({'acc': batch4_metrics, 'y': batch4_labels}),
            Data({'acc': batch5_metrics, 'y': batch5_labels}),
            Data({'acc': batch6_metrics, 'y': batch6_labels})
            ]]

        cls.epoch_1_means = {0: 0.49, 1: 0.42, 2: 0.59}
        cls.epoch_2_means = {0: 0.54, 1: 0.66, 2: 0.45}

    def _simulate_training(self, trace: LabelTracker, data: Data):
        trace.on_begin(Data())
        system = trace.system
        global_step = 0
        for epoch_idx, epoch in enumerate(self.training_data, start=1):
            system.epoch_idx = epoch_idx
            trace.on_epoch_begin(Data())
            for batch_idx, batch in enumerate(epoch, start=1):
                system.batch_idx = batch_idx
                global_step += 1
                system.global_step = global_step
                trace.on_batch_begin(Data())
                trace.on_batch_end(batch)
            trace.on_epoch_end(Data())
        trace.on_end(data)

    def test_basic_happy_path(self):
        labeltracker = LabelTracker(label='y', metric='acc', bounds=None, outputs='out')
        system = sample_system_object()
        labeltracker.system = system
        response = Data()
        self._simulate_training(labeltracker, response)

        # Check responses
        with self.subTest('Check that a response was written'):
            self.assertIn('out', response)
        with self.subTest('Check that data was written to the system'):
            self.assertIn('out', system.custom_graphs)
        response = response['out']
        with self.subTest('Check consistency of outputs'):
            self.assertEqual(response, system.custom_graphs['out'])
        with self.subTest('Check that all 3 labels have summaries'):
            self.assertEqual(3, len(response))
        with self.subTest('Check correct mean values (epoch 1)'):
            self.assertDictEqual(self.epoch_1_means,
                                 {elem.name: round(elem.history['train']['acc'][3], 6)
                                  for elem in response})
        with self.subTest('Check correct mean values (epoch 2)'):
            self.assertDictEqual(self.epoch_2_means,
                                 {elem.name: round(elem.history['train']['acc'][6], 6)
                                  for elem in response})

    def test_basic_limiting_labels(self):
        labeltracker = LabelTracker(label='y',
                                    metric='acc',
                                    bounds=None,
                                    label_mapping={'good': 0, 'bad': 1},
                                    outputs='out')
        system = sample_system_object()
        labeltracker.system = system
        response = Data()
        self._simulate_training(labeltracker, response)

        # Check responses
        with self.subTest('Check that a response was written'):
            self.assertIn('out', response)
        with self.subTest('Check that data was written to the system'):
            self.assertIn('out', system.custom_graphs)
        response = response['out']
        with self.subTest('Check consistency of outputs'):
            self.assertEqual(response, system.custom_graphs['out'])
        with self.subTest('Check that all 2 labels have summaries'):
            self.assertEqual(2, len(response))
        with self.subTest('Check correct mean values (epoch 1)'):
            target = {'good': self.epoch_1_means[0], 'bad': self.epoch_1_means[1]}
            self.assertDictEqual(target,
                                 {elem.name: round(elem.history['train']['acc'][3], 6)
                                  for elem in response})
        with self.subTest('Check correct mean values (epoch 2)'):
            target = {'good': self.epoch_2_means[0], 'bad': self.epoch_2_means[1]}
            self.assertDictEqual(target,
                                 {elem.name: round(elem.history['train']['acc'][6], 6)
                                  for elem in response})

    def test_multiple_bounds(self):
        labeltracker = LabelTracker(label='y', metric='acc', bounds=['std', 'range'], outputs='out')
        system = sample_system_object()
        labeltracker.system = system
        response = Data()
        self._simulate_training(labeltracker, response)

        # Check responses
        with self.subTest('Check that a response was written'):
            self.assertIn('out', response)
        with self.subTest('Check that data was written to the system'):
            self.assertIn('out', system.custom_graphs)
        response = response['out']
        with self.subTest('Check consistency of outputs'):
            self.assertEqual(response, system.custom_graphs['out'])
        with self.subTest('Check that all 3 labels have summaries'):
            self.assertEqual(3, len(response))
        with self.subTest('Check that regular mean is not present'):
            for elem in response:
                self.assertNotIn('acc', elem.history['train'])
        with self.subTest('Check that stddev and range are both present'):
            for elem in response:
                self.assertIn('acc ($\\mu \\pm \\sigma$)', elem.history['train'])
                self.assertIn('acc ($min, \\mu, max$)', elem.history['train'])

    def test_save_and_load_state(self):
        def instantiate_system():
            tracker = LabelTracker(label='y', metric='acc', bounds=[None, 'range'], outputs='out')
            system = sample_system_object()
            system.traces.append(tracker)
            tracker.system = system
            return system, tracker

        system, tracker = instantiate_system()

        # Make some changes
        response = Data()
        self._simulate_training(tracker, response)

        # Save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_dir=save_path)

        # reinstantiate system and load the state
        system, tracker = instantiate_system()
        system.load_state(save_path)

        loaded_tracker = system.traces[-1]
        response = list(loaded_tracker.label_summaries.values())
        with self.subTest('Check that all 3 labels have summaries'):
            self.assertEqual(3, len(response))
        with self.subTest('Check correct mean values (epoch 1)'):
            self.assertDictEqual(self.epoch_1_means,
                                 {elem.name: round(elem.history['train']['acc'][3], 6)
                                  for elem in response})
        with self.subTest('Check correct mean values (epoch 2)'):
            self.assertDictEqual(self.epoch_2_means,
                                 {elem.name: round(elem.history['train']['acc'][6], 6)
                                  for elem in response})



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict, Iterable, List, Optional, Union

import numpy as np
from scipy.linalg import hadamard

from fastestimator.backend.gather import gather
from fastestimator.op.numpyop.numpyop import NumpyOp
from fastestimator.util.traceability_util import traceable


@traceable()
class Hadamard(NumpyOp):
    """Convert integer labels into hadamard code representations.

    Args:
        inputs: Key of the input tensor(s) to be converted.
        outputs: Key of the output tensor(s) in hadamard code representation.
        n_classes: How many classes are there in the inputs.
        code_length: What code length to use. Will default to the smallest power of 2 which is >= the number of classes.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".

    Raises:
        ValueError: If an unequal number of `inputs` and `outputs` are provided, or if `code_length` is invalid.
    """
    def __init__(self,
                 inputs: Union[str, List[str]],
                 outputs: Union[str, List[str]],
                 n_classes: int,
                 code_length: Optional[int] = None,
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None) -> None:
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        self.in_list, self.out_list = True, True
        if len(self.inputs) != len(self.outputs):
            raise ValueError("Hadamard requires the same number of input and output keys.")
        self.n_classes = n_classes
        if code_length is None:
            code_length = 1 << (n_classes - 1).bit_length()
        if code_length <= 0 or (code_length & (code_length - 1) != 0):
            raise ValueError(f"code_length must be a positive power of 2, but got {code_length}.")
        if code_length < n_classes:
            raise ValueError(f"code_length must be >= n_classes, but got {code_length} and {n_classes}.")
        self.code_length = code_length
        labels = hadamard(self.code_length).astype(np.float32)
        labels[np.arange(0, self.code_length, 2), 0] = -1  # Make first column alternate
        labels = labels[:self.n_classes]
        self.labels = labels

    def forward(self, data: List[Union[int, np.ndarray]], state: Dict[str, Any]) -> List[np.ndarray]:
        # TODO - also support one hot with smoothed labels?
        return [gather(tensor=self.labels, indices=np.array(inp)) for inp in data]



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe
import fastestimator.test.unittest_util as fet


class TestToTensor(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data_np = {
            "x": np.ones((10, 15), dtype=np.int32),
            "y": [np.ones((4), dtype=np.float32), np.ones((5, 3), dtype=np.float32)],
            "z": {
                "key": np.ones((2, 2), dtype=np.float64)
            }
        }

        cls.data_tf = {
            "x": tf.ones((10, 15), dtype=tf.int32),
            "y": [tf.ones((4), dtype=tf.float32), tf.ones((5, 3), dtype=tf.float32)],
            "z": {
                "key": tf.ones((2, 2), dtype=tf.float64)
            }
        }
        cls.data_torch = {
            "x": torch.ones((10, 15), dtype=torch.int32),
            "y": [torch.ones((4), dtype=torch.float32), torch.ones((5, 3), dtype=torch.float32)],
            "z": {
                "key": torch.ones((2, 2), dtype=torch.float64)
            }
        }

    def test_to_tensor_np_to_tf(self):
        self.assertTrue(
            fet.is_equal(fe.backend.to_tensor(self.data_np, target_type='tf'), self.data_tf, assert_dtype=True))

    def test_to_tensor_np_to_torch(self):
        self.assertTrue(
            fet.is_equal(fe.backend.to_tensor(self.data_np, target_type='torch'), self.data_torch, assert_dtype=True))

    def test_to_tensor_tf_to_torch(self):
        self.assertTrue(
            fet.is_equal(fe.backend.to_tensor(self.data_tf, target_type='torch'), self.data_torch, assert_dtype=True))

    def test_to_tensor_torch_to_tf(self):
        self.assertTrue(
            fet.is_equal(fe.backend.to_tensor(self.data_torch, target_type='tf'), self.data_tf, assert_dtype=True))

    def test_tf_with_nones(self):
        n = np.ones((13, 3, 2, 4), dtype=np.float32)
        t = tf.ones((13, 3, 2, 4), dtype=tf.float32)
        self.assertTrue(fet.is_equal(fe.backend.to_tensor([n, None], target_type='tf'), [t, None], assert_dtype=True))

    def test_torch_with_nones(self):
        n = np.ones((13, 3, 2, 4), dtype=np.int16)
        p = torch.ones((13, 3, 2, 4), dtype=torch.int16)
        self.assertTrue(fet.is_equal(fe.backend.to_tensor((None, n), target_type='torch'), (None, p),
                                     assert_dtype=True))
